import pytest
from rlberry.agents.kernel_based import RSKernelUCBVIAgent
from rlberry.agents.kernel_based import RSUCBVIAgent
from rlberry.envs.benchmarks.ball_exploration.ball2d import get_benchmark_env


@pytest.mark.parametrize("kernel_type", [
                "uniform",
                "triangular",
                "gaussian",
                "epanechnikov",
                "quartic",
                "triweight",
                "tricube",
                "cosine"
])
def test_rs_kernel_ucbvi(kernel_type):
    for horizon in [None, 30]:
        env = get_benchmark_env(level=1)
        agent = RSKernelUCBVIAgent(
                                env,
                                n_episodes=5,
                                gamma=0.95,
                                horizon=horizon,
                                bonus_scale_factor=0.01,
                                min_dist=0.2,
                                bandwidth=0.05,
                                beta=1.0,
                                kernel_type=kernel_type,
                                verbose=1)
        agent._log_interval = 0
        agent.fit()
        agent.policy(env.observation_space.sample())


def test_rs_ucbvi():
    env = get_benchmark_env(level=1)
    agent = RSUCBVIAgent(env,
                         n_episodes=5,
                         gamma=0.99,
                         horizon=30,
                         bonus_scale_factor=0.1,
                         verbose=1)
    agent._log_interval = 0
    agent.fit()
    agent.policy(env.observation_space.sample())



import rlberry.seeding as seeding
import logging

from rlberry.envs.interface import Model
from rlberry.wrappers.gym_utils import convert_space_from_gym


_GYM_INSTALLED = True
try:
    import gym
except Exception:
    _GYM_INSTALLED = False


class Wrapper(Model):
    """
    Wraps a given environment, similar to OpenAI gym's wrapper [1].
    Can also be used to wrap gym environments.

    Note:
        The input environment is not copied (Wrapper.env points
        to the input env).

    See also:
    https://stackoverflow.com/questions/1443129/completely-wrap-an-object-in-python

    [1] https://github.com/openai/gym/blob/master/gym/core.py
    """
    def __init__(self, env):
        # Init base class
        Model.__init__(self)

        # Save reference to env
        self.env = env

        # Check if gym environment
        if _GYM_INSTALLED and isinstance(env, gym.Env):
            gym_env = env
            # Warnings
            logging.warning(
                'GymWrapper: Rendering gym.Env does not \
follow the same protocol as rlberry.')
            # Convert spaces
            self.observation_space = convert_space_from_gym(
                gym_env.observation_space)
            self.action_space = convert_space_from_gym(gym_env.action_space)
            # Reward range
            self.reward_range = gym_env.reward_range

        # If rlberry environment
        else:
            self.observation_space = self.env.observation_space
            self.action_space = self.env.action_space
            self.reward_range = self.env.reward_range

    @property
    def unwrapped(self):
        return self.env.unwrapped

    def __getattr__(self, attr):
        """
        The first condition is to avoid infinite recursion when deep copying.
        See https://stackoverflow.com/a/47300262
        """
        if attr[:2] == '__':
            raise AttributeError(attr)
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self.env, attr)

    def reseed(self):
        self.rng = seeding.get_rng()
        if _GYM_INSTALLED and isinstance(self.env, gym.Env):
            # get a seed for gym environment
            self.env.seed(self.rng.integers(2**16).item())
        else:
            self.env.reseed()

    def reset(self):
        return self.env.reset()

    def step(self, action):
        return self.env.step(action)

    def sample(self, state, action):
        return self.env.sample(state, action)

    def is_online(self):
        try:
            self.env.reset()
            self.env.step(self.env.action_space.sample())
            return True
        except Exception:
            return False

    def is_generative(self):
        try:
            self.env.sample(self.env.observation_space.sample(),
                            self.env.action_space.sample())
            return True
        except Exception:
            return False

    def __str__(self):
        return '<{}{}>'.format(type(self).__name__, self.env)



import numpy as np

import rlberry.spaces as spaces
from rlberry.utils.binsearch import binary_search_nd, unravel_index_uniform_bin
from rlberry.wrappers import Wrapper


class DiscretizeStateWrapper(Wrapper):
    """
    Discretize an environment with continuous states and discrete actions.
    """

    def __init__(self, _env, n_bins):
        # initialize base class
        super().__init__(_env)

        self.n_bins = n_bins
        # initialize bins
        assert n_bins > 0, "DiscretizeStateWrapper requires n_bins > 0"
        n_states = 1
        tol = 1e-8
        self.dim = len(self.env.observation_space.low)
        n_states = n_bins ** self.dim
        self._bins = []
        self._open_bins = []
        for dd in range(self.dim):
            range_dd = self.env.observation_space.high[dd] \
                - self.env.observation_space.low[dd]
            epsilon = range_dd / n_bins
            bins_dd = []
            for bb in range(n_bins + 1):
                val = self.env.observation_space.low[dd] + epsilon * bb
                bins_dd.append(val)
            self._open_bins.append(tuple(bins_dd[1:]))
            bins_dd[-1] += tol  # "close" the last interval
            self._bins.append(tuple(bins_dd))

            # set observation space
        self.observation_space = spaces.Discrete(n_states)

        # List of discretized states
        self.discretized_states = np.zeros((self.dim, n_states))
        for ii in range(n_states):
            self.discretized_states[:, ii] = \
                self.get_continuous_state(ii, False)

    def reset(self):
        return self.get_discrete_state(self.env.reset())

    def step(self, action):
        next_state, reward, done, info = self.env.step(action)
        next_state = binary_search_nd(next_state, self._bins)
        return next_state, reward, done, info

    def sample(self, discrete_state, action):
        # map disctete state to continuous one
        assert self.observation_space.contains(discrete_state)
        continuous_state = self.get_continuous_state(discrete_state,
                                                     randomize=True)
        # sample in the true environment
        next_state, reward, done, info = \
            self.env.sample(continuous_state, action)
        # discretize next state
        next_state = binary_search_nd(next_state, self._bins)

        return next_state, reward, done, info

    def get_discrete_state(self, continuous_state):
        return binary_search_nd(continuous_state, self._bins)

    def get_continuous_state(self, discrete_state, randomize=False):
        assert discrete_state >= 0 \
            and discrete_state < self.observation_space.n, \
            "invalid discrete_state"
        # get multi-index
        index \
            = unravel_index_uniform_bin(discrete_state, self.dim, self.n_bins)

        # get state
        continuous_state = np.zeros(self.dim)
        for dd in range(self.dim):
            continuous_state[dd] = self._bins[dd][index[dd]]
            if randomize:
                range_dd = self.env.observation_space.high[dd] \
                    - self.env.observation_space.low[dd]
                epsilon = range_dd / self.n_bins
                continuous_state[dd] += epsilon * self.rng.uniform()
        return continuous_state



import logging
import numpy as np
from rlberry.utils.jit_setup import numba_jit
from rlberry.exploration_tools.uncertainty_estimator \
    import UncertaintyEstimator
from rlberry.spaces import Box, Discrete
from rlberry.utils.metrics import metric_lp


@numba_jit
def map_to_representative(state,
                          lp_metric,
                          representative_states,
                          n_representatives,
                          min_dist,
                          scaling,
                          accept_new_repr):
    """
    Map state to representative state.
    """
    dist_to_closest = np.inf
    argmin = -1
    for ii in range(n_representatives):
        dist = metric_lp(state, representative_states[ii, :],
                         lp_metric, scaling)
        if dist < dist_to_closest:
            dist_to_closest = dist
            argmin = ii

    max_representatives = representative_states.shape[0]
    if dist_to_closest > min_dist \
        and n_representatives < max_representatives \
            and accept_new_repr:
        new_index = n_representatives
        representative_states[new_index, :] = state
        return new_index, 0.0
    return argmin, dist_to_closest


class OnlineDiscretizationCounter(UncertaintyEstimator):
    """
    Note: currently, only implemented for continuous (Box) states and
    discrete actions.
    """
    def __init__(self,
                 observation_space,
                 action_space,
                 lp_metric=2,
                 min_dist=0.1,
                 max_repr=1000,
                 scaling=None,
                 **kwargs):
        """
        Parameters
        ----------
        observation_space : spaces.Box
        action_space : spaces.Discrete
        lp_metric: int
            The metric on the state space is the one induced by the p-norm,
            where p = lp_metric. Default = 2, for the Euclidean metric.
        scaling: numpy.ndarray
            Must have the same size as state array, used to scale the states
            before computing the metric.
            If None, set to:
            - (env.observation_space.high - env.observation_space.low) if high
            and low are bounded
            - np.ones(env.observation_space.dim) if high or low are unbounded
        min_dist: double
            Minimum distance between two representative states
        max_repr: int
            Maximum number of representative states.
            If None, it is set to  (sqrt(d)/min_dist)**d, where d
            is the dimension of the state space
        """
        UncertaintyEstimator.__init__(self, observation_space, action_space)

        assert isinstance(action_space, Discrete)
        assert isinstance(observation_space, Box)

        self.lp_metric = lp_metric
        self.min_dist = min_dist
        self.max_repr = max_repr
        self.state_dim = self.observation_space.dim
        self.n_actions = self.action_space.n

        # compute scaling, if it is None
        if scaling is None:
            # if high and low are bounded
            if self.observation_space.is_bounded:
                scaling = self.observation_space.high \
                    - self.observation_space.low
                # if high or low are unbounded
            else:
                scaling = np.ones(self.state_dim)
        else:
            assert scaling.ndim == 1
            assert scaling.shape[0] == self.state_dim
        self.scaling = scaling

        # initialize
        self.n_representatives = None
        self.representative_states = None
        self.N_sa = None
        self.reset()

    def reset(self):
        self.n_representatives = 0
        self.representative_states = np.zeros((self.max_repr, self.state_dim))
        self.N_sa = np.zeros((self.max_repr, self.n_actions))

        self._overflow_warning = False

    def _get_representative_state(self, state, accept_new_repr=True):
        state_idx, dist_to_closest \
            = map_to_representative(state,
                                    self.lp_metric,
                                    self.representative_states,
                                    self.n_representatives,
                                    self.min_dist,
                                    self.scaling,
                                    accept_new_repr)
        # check if new representative state
        if state_idx == self.n_representatives:
            self.n_representatives += 1

        if self.n_representatives >= self.max_repr \
                and (not self._overflow_warning):
            logging.warning("OnlineDiscretizationCounter reached \
the maximum number of representative states.")
            self._overflow_warning = True

        return state_idx, dist_to_closest

    def update(self, state, action, next_state, reward, **kwargs):
        state_idx, _ = self._get_representative_state(state)
        self.N_sa[state_idx, action] += 1

    def measure(self, state, action, **kwargs):
        n = np.maximum(1.0, self.count(state, action))
        return 1.0/np.sqrt(n)

    def count(self, state, action):
        state_idx, dist_to_closest = self._get_representative_state(
                                            state,
                                            accept_new_repr=False)
        # if state is too far from the closest representative,
        # its count is zero.
        if dist_to_closest > self.min_dist:
            return 0.0
        return self.N_sa[state_idx, action]



from rlberry.envs import GridWorld
from rlberry.envs import MountainCar
from rlberry.exploration_tools.discrete_counter import DiscreteCounter
from rlberry.exploration_tools.online_discretization_counter \
    import OnlineDiscretizationCounter


def test_discrete_env():
    env = GridWorld()
    counter = DiscreteCounter(env.observation_space, env.action_space)

    for N in range(10, 20):
        for ss in range(env.observation_space.n):
            for aa in range(env.action_space.n):
                for _ in range(N):
                    ns, rr, _, _ = env.sample(ss, aa)
                    counter.update(ss, aa, ns, rr)
                assert counter.N_sa[ss, aa] == N
                assert counter.count(ss, aa) == N
        counter.reset()


def test_continuous_state_env():
    env = MountainCar()
    counter = DiscreteCounter(env.observation_space, env.action_space)

    for N in [10, 20, 30]:
        for _ in range(100):
            ss = env.observation_space.sample()
            aa = env.action_space.sample()
            for _ in range(N):
                ns, rr, _, _ = env.sample(ss, aa)
                counter.update(ss, aa, ns, rr)

            dss = counter.state_discretizer.discretize(ss)
            assert counter.N_sa[dss, aa] == N
            assert counter.count(ss, aa) == N
            counter.reset()


def test_continuous_state_env_2():
    env = MountainCar()
    counter = OnlineDiscretizationCounter(env.observation_space,
                                          env.action_space)

    for N in [10, 20, 30]:
        for _ in range(100):
            ss = env.observation_space.sample()
            aa = env.action_space.sample()
            for nn in range(N):
                ns, rr, _, _ = env.sample(ss, aa)
                counter.update(ss, aa, ns, rr)
            assert counter.count(ss, aa) == N
            counter.reset()




from rlberry.agents.cem import CEMAgent
from rlberry.envs.benchmarks.ball_exploration.ball2d import get_benchmark_env


def test_cem_agent():
    env = get_benchmark_env(level=1)
    n_episodes = 5
    batch_size = 100
    horizon = 30
    gamma = 0.99

    agent = CEMAgent(env,
                     n_episodes,
                     horizon,
                     gamma,
                     batch_size,
                     percentile=70,
                     learning_rate=0.01,
                     verbose=1)
    agent._log_interval = 0
    agent.fit()
    agent.policy(env.observation_space.sample())



import numpy as np
from rlberry.utils.jit_setup import numba_jit


@numba_jit
def metric_lp(x, y, p, scaling):
    """
    Returns the p-norm:  || (x-y)/scaling||_p

    Parameters
    ----------
    x : numpy.ndarray
        1d array
    y : numpy.ndarray
        1d array
    p : int
        norm parameter
    scaling : numpy.ndarray
        1d array
    """
    assert p >= 1
    assert x.ndim == 1
    assert y.ndim == 1
    assert scaling.ndim == 1

    d = len(x)
    diff = np.abs((x - y) / scaling)
    # p = infinity
    if p == np.inf:
        return diff.max()
    # p < infinity
    tmp = 0
    for ii in range(d):
        tmp += np.power(diff[ii], p)
    return np.power(tmp, 1.0 / p)



import numpy as np

import rlberry.spaces as spaces
from rlberry.envs.interface import Model


class FiniteMDP(Model):
    """
    Base class for a finite MDP.

    Attributes
    ----------
    R : numpy.ndarray
        array of shape (S, A) containing the mean rewards, where
        S = number of states;  A = number of actions.
    P : numpy.ndarray
        array of shape (S, A, S) containing the transition probabilities,
        where P[s, a, s'] = Prob(S_{t+1}=s'| S_t = s, A_t = a).
    """

    def __init__(self, R, P, initial_state_distribution=0):
        """
        Parameters
        ----------
        R : numpy.ndarray
        P : numpy.ndarray
        initial_state_distribution : numpy.ndarray or int
            array of size (S,) containing the initial state distribution
            or an integer representing the initial/default state
        """
        Model.__init__(self)
        self.initial_state_distribution = initial_state_distribution
        S, A = R.shape

        self.S = S
        self.A = A

        self.R = R
        self.P = P

        self.observation_space = spaces.Discrete(S)
        self.action_space = spaces.Discrete(A)
        self.reward_range = (self.R.min(), self.R.max())

        self.state = None

        self._states = np.arange(S)
        self._actions = np.arange(A)

        self.reset()
        self._check()

    def reset(self):
        """
        Reset the environment to a default state.
        """
        if isinstance(self.initial_state_distribution, np.ndarray):
            self.state = self.rng.choice(self._states,
                                         p=self.initial_state_distribution)
        else:
            self.state = self.initial_state_distribution
        return self.state

    def _check(self):
        """
        Check consistency of the MDP
        """
        # Check initial_state_distribution
        if isinstance(self.initial_state_distribution, np.ndarray):
            assert abs(self.initial_state_distribution.sum() - 1.0) < 1e-15
        else:
            assert self.initial_state_distribution >= 0
            assert self.initial_state_distribution < self.S

        # Check that P[s,a, :] is a probability distribution
        for s in self._states:
            for a in self._actions:
                assert abs(self.P[s, a, :].sum() - 1.0) < 1e-15

        # Check that dimensions match
        S1, A1 = self.R.shape
        S2, A2, S3 = self.P.shape
        assert S1 == S2 == S3
        assert A1 == A2

    def sample(self, state, action):
        """
        Sample a transition s' from P(s'|state, action).
        """
        prob = self.P[state, action, :]
        next_state = self.rng.choice(self._states, p=prob)
        reward = self.reward_fn(state, action, next_state)
        done = self.is_terminal(self.state)
        info = {}
        return next_state, reward, done, info

    def step(self, action):
        assert action in self._actions, "Invalid action!"
        next_state, reward, done, info = self.sample(self.state, action)
        self.state = next_state
        return next_state, reward, done, info

    def is_terminal(self, state):
        """
        Returns true if a state is terminal.
        """
        return False

    def reward_fn(self, state, action, next_state):
        """
        Reward function. Returns mean reward at (state, action) by default.

        Parameters
        ----------
        state : int
            current state
        action : int
            current action
        next_state :
            next state

        Returns:
            reward : float
        """
        return self.R[state, action]

    def print(self):
        """
        Print the structure of the MDP.
        """
        indent = '    '
        for s in self._states:
            print(("State %d" + indent) % s)
            for a in self._actions:
                print(indent + "Action ", a)
                for ss in self._states:
                    if self.P[s, a, ss] > 0.0:
                        print(2 * indent + 'transition to %d with prob %0.2f'
                              % (ss, self.P[s, a, ss]))
            print("~~~~~~~~~~~~~~~~~~~~")


# if __name__ == '__main__':
#     S = 3
#     A = 2

#     R = np.random.uniform(0, 1, (S, A))
#     P = np.random.uniform(0, 1, (S, A, S))
#     initial_state_distr = 1  # np.ones(S)/S
#     for ss in range(S):
#         for aa in range(A):
#             P[ss, aa, :] /= P[ss, aa, :].sum()

#     env = FiniteMDP(R, P, initial_state_distribution=initial_state_distr)



#
# Checks if numba is installed.
# -> If so, use numba.jit
# -> Otherwise, define numba_jit as a dummy decorator.
#

try:
    from numba import jit
    numba_jit = jit(nopython=True)
except Exception:
    def numba_jit(func, **options):
        """This decorator does not modify the decorated function."""
        return func



import numpy as np
import pytest

from rlberry.envs.classic_control import MountainCar, Acrobot
from rlberry.envs.finite import Chain
from rlberry.envs.finite import GridWorld
from rlberry.envs.benchmarks.ball_exploration import PBall2D, SimplePBallND
from rlberry.envs.benchmarks.ball_exploration.ball2d import get_benchmark_env
from rlberry.rendering.render_interface import RenderInterface2D


classes = [
    MountainCar,
    GridWorld,
    Chain,
    PBall2D,
    SimplePBallND,
    Acrobot
]


@pytest.mark.parametrize("ModelClass", classes)
def test_instantiation(ModelClass):
    env = ModelClass()

    if env.is_online():
        for _ in range(2):
            state = env.reset()
            for _ in range(50):
                assert env.observation_space.contains(state)
                action = env.action_space.sample()
                next_s, _, _, _ = env.step(action)
                state = next_s

    if env.is_generative():
        for _ in range(100):
            state = env.observation_space.sample()
            action = env.action_space.sample()
            next_s, _, _, _ = env.sample(state, action)
            assert env.observation_space.contains(next_s)


@pytest.mark.parametrize("ModelClass", classes)
def test_rendering_calls(ModelClass):
    env = ModelClass()
    if isinstance(env, RenderInterface2D):
        _ = env.get_background()
        _ = env.get_scene(env.observation_space.sample())


def test_gridworld_aux_functions():
    env = GridWorld(nrows=5, ncols=5, walls=((1, 1),),
                    reward_at={(4, 4): 1, (4, 3): -1})
    env.print()  # from FiniteMDP
    env.render_ascii()  # from GridWorld
    vals = np.ones(env.observation_space.n)
    env.display_values(vals)
    env.print_transition_at(0, 0, 'up')


def test_ball2d_benchmark_instantiation():
    for level in [1, 2, 3, 4, 5]:
        env = get_benchmark_env(level)
        for aa in range(env.action_space.n):
            env.step(aa)
            env.sample(env.observation_space.sample(), aa)


@pytest.mark.parametrize("p", [1, 2, 3, 4, 5, np.inf])
def test_pball_env(p):
    env = PBall2D(p=p)
    env.get_reward_lipschitz_constant()
    env.get_transitions_lipschitz_constant()



import numpy as np
from rlberry.spaces import Box, Discrete
from rlberry.utils.binsearch import binary_search_nd
from rlberry.utils.binsearch import unravel_index_uniform_bin


class Discretizer:
    def __init__(self, space, n_bins):
        assert isinstance(space, Box), \
            "Discretization is only implemented for Box spaces."
        assert space.is_bounded
        self.space = space
        self.n_bins = n_bins

        # initialize bins
        assert n_bins > 0, "Discretizer requires n_bins > 0"
        n_elements = 1
        tol = 1e-8
        self.dim = len(self.space.low)
        n_elements = n_bins ** self.dim
        self._bins = []
        self._open_bins = []
        for dd in range(self.dim):
            range_dd = self.space.high[dd] - self.space.low[dd]
            epsilon = range_dd / n_bins
            bins_dd = []
            for bb in range(n_bins + 1):
                val = self.space.low[dd] + epsilon * bb
                bins_dd.append(val)
            self._open_bins.append(tuple(bins_dd[1:]))
            bins_dd[-1] += tol  # "close" the last interval
            self._bins.append(tuple(bins_dd))

        # set observation space
        self.discrete_space = Discrete(n_elements)

        # List of discretized elements
        self.discretized_elements = np.zeros((self.dim, n_elements))
        for ii in range(n_elements):
            self.discretized_elements[:, ii] = self.get_coordinates(ii, False)

    def discretize(self, coordinates):
        return binary_search_nd(coordinates, self._bins)

    def get_coordinates(self, flat_index, randomize=False):
        assert self.discrete_space.contains(flat_index), "invalid flat_index"
        # get multi-index
        index = unravel_index_uniform_bin(flat_index, self.dim, self.n_bins)

        # get coordinates
        coordinates = np.zeros(self.dim)
        for dd in range(self.dim):
            coordinates[dd] = self._bins[dd][index[dd]]
            if randomize:
                range_dd = self.space.high[dd] - self.space.low[dd]
                epsilon = range_dd / self.n_bins
                coordinates[dd] += epsilon * self.space.rng.uniform()
        return coordinates



from rlberry.envs.interface import Model
from rlberry.wrappers import Wrapper
from rlberry.envs import GridWorld
from rlberry.spaces import Discrete, Box

_GYM_INSTALLED = True
try:
    import gym
except Exception:
    _GYM_INSTALLED = False


def test_wrapper():
    global _GYM_INSTALLED

    env = GridWorld()
    wrapped = Wrapper(env)
    assert isinstance(wrapped, Model)
    assert wrapped.is_online()
    assert wrapped.is_generative()

    # calling some functions
    wrapped.reset()
    wrapped.step(wrapped.action_space.sample())
    wrapped.sample(wrapped.observation_space.sample(),
                   wrapped.action_space.sample())


def test_gym_wrapper():
    if _GYM_INSTALLED:
        gym_env = gym.make('Acrobot-v1')
        wrapped = Wrapper(gym_env)
        assert isinstance(wrapped, Model)
        assert wrapped.is_online()
        assert not wrapped.is_generative()

        assert isinstance(wrapped.observation_space, Box)
        assert isinstance(wrapped.action_space, Discrete)

        # calling some gym functions
        wrapped.close()
        wrapped.seed()



import logging
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

import rlberry.seeding as seeding


def plot_episode_rewards(agent_stats_list, cumulative=False,
                         fignum=None, show=True):
    plt.figure(fignum)
    for agent_stats in agent_stats_list:
        if 'episode_rewards' not in agent_stats.fit_info:
            logging.warning("episode_rewards not \
available for %s." % agent_stats.agent_name)
            continue
        else:
            # train agents if they are not already trained
            if agent_stats.fitted_agents is None:
                agent_stats.fit()
            # get reward statistics and plot them
            rewards = np.array(agent_stats.fit_statistics['episode_rewards'])
            if cumulative:
                rewards = np.cumsum(rewards, axis=1)
            mean_r = rewards.mean(axis=0)
            std_r = rewards.std(axis=0)
            episodes = np.arange(1, rewards.shape[1]+1)

            plt.plot(episodes, mean_r, label=agent_stats.agent_name)
            plt.fill_between(episodes, mean_r-std_r, mean_r+std_r, alpha=0.2)
            plt.legend()
            plt.xlabel("episodes")
            if not cumulative:
                plt.ylabel("reward in one episode")
            else:
                plt.ylabel("total reward")
            plt.grid(True, alpha=0.75)

    if show:
        plt.show()


def compare_policies(agent_stats_list, eval_env=None, eval_horizon=None,
                     stationary_policy=True, n_sim=10, fignum=None,
                     show=True, plot=True):
    """
    Compare the policies of each of the agents in agent_stats_list.
    Each element of the agent_stats_list contains a list of fitted agents.
    To evaluate the policy, we repeat n_sim times:
        * choose one of the fitted agents uniformly at random
        * run its policy in eval_env for eval_horizon time steps

    To do
    ------
    Paralellize evaluations of each agent.

    Parameters
    ----------
    agent_stats_list : list of AgentStats objects.
    eval_env : Model
        Environment where to evaluate the policies.
        If None, it is taken from AgentStats.
    eval_horizon : int
        Number of time steps for policy evaluation.
        If None, it is taken from AgentStats.
    stationary_policy : bool
        If False, the time step h (0<= h <= eval_horizon) is sent as input
        to agent.policy() for policy evaluation.
    n_sim : int
        Number of simulations to evaluate each policy.
    fignum: string or int
        Identifier of plot figure
    show: bool
        If true, calls plt.show()
    plot: bool
        If false, do not plot.
    """
    #
    # evaluation
    #
    use_eval_from_agent_stats = (eval_env is None)
    use_horizon_from_agent_stats = (eval_horizon is None)

    rng = seeding.get_rng()
    agents_rewards = []
    for agent_stats in agent_stats_list:
        # train agents if they are not already trained
        if agent_stats.fitted_agents is None:
            agent_stats.fit()

        # eval env and horizon
        if use_eval_from_agent_stats:
            eval_env = agent_stats.eval_env
            assert eval_env is not None, \
                "eval_env not in AgentStats %s" % agent_stats.agent_name
        if use_horizon_from_agent_stats:
            eval_horizon = agent_stats.eval_horizon
            assert eval_horizon is not None, \
                "eval_horizon not in AgentStats %s" % agent_stats.agent_name

        # evaluate agent
        episode_rewards = np.zeros(n_sim)
        for sim in range(n_sim):
            # choose one of the fitted agents randomly
            agent_idx = rng.integers(len(agent_stats.fitted_agents))
            agent = agent_stats.fitted_agents[agent_idx]
            # evaluate agent
            observation = eval_env.reset()
            for hh in range(eval_horizon):
                if stationary_policy:
                    action = agent.policy(observation,
                                          **agent_stats.policy_kwargs)
                else:
                    action = agent.policy(observation, hh,
                                          **agent_stats.policy_kwargs)
                observation, reward, done, _ = eval_env.step(action)
                episode_rewards[sim] += reward
                if done:
                    break
        # store rewards
        agents_rewards.append(episode_rewards)

    #
    # plot
    #

    # build unique agent IDs (in case there are two agents with the same ID)
    unique_ids = []
    id_count = {}
    for agent_stats in agent_stats_list:
        name = agent_stats.agent_name
        if name not in id_count:
            id_count[name] = 1
        else:
            id_count[name] += 1

        unique_ids.append(name + "*"*(id_count[name]-1))

    # convert output to DataFrame
    data = {}
    for agent_id, agent_rewards in zip(unique_ids, agents_rewards):
        data[agent_id] = agent_rewards
    output = pd.DataFrame(data)

    # plot
    if plot:
        plt.figure(fignum)

        with sns.axes_style("whitegrid"):
            ax = sns.boxplot(data=output)
            ax.set_xlabel("agent")
            ax.set_ylabel("rewards in one episode")
            plt.title("Environment = %s" % eval_env.unwrapped.name)
            if show:
                plt.show()

    return output



import numpy as np
from rlberry.rendering import GeometricPrimitive


def bar_shape(p0, p1, width):
    shape = GeometricPrimitive("QUADS")

    x0, y0 = p0
    x1, y1 = p1

    direction = np.array([x1-x0, y1-y0])
    norm = np.sqrt((direction*direction).sum())
    direction = direction/norm

    # get vector perpendicular to direction
    u_vec = np.zeros(2)
    u_vec[0] = -direction[1]
    u_vec[1] = direction[0]

    u_vec = u_vec*width/2

    shape.add_vertex((x0+u_vec[0], y0+u_vec[1]))
    shape.add_vertex((x0-u_vec[0], y0-u_vec[1]))
    shape.add_vertex((x1-u_vec[0], y1-u_vec[1]))
    shape.add_vertex((x1+u_vec[0], y1+u_vec[1]))
    return shape


def circle_shape(center, radius, n_points=50):
    shape = GeometricPrimitive("POLYGON")

    x0, y0 = center
    theta = np.linspace(0.0, 2*np.pi, n_points)
    for tt in theta:
        xx = radius*np.cos(tt)
        yy = radius*np.sin(tt)
        shape.add_vertex((x0+xx, y0+yy))

    return shape



"""
Code for 2D rendering, using pygame (without OpenGL)
"""

import numpy as np
from os import environ
from rlberry.rendering import Scene

environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'

_IMPORT_SUCESSFUL = True
try:
    import pygame as pg

except Exception:
    _IMPORT_SUCESSFUL = False


class PyGameRender2D:
    """Class to render a list of scenes using pygame."""

    def __init__(self):
        # parameters
        self.window_width = 800
        self.window_height = 800    # multiples of 16 are preferred
        self.background_color = (150, 190, 255)
        self.refresh_interval = 50
        self.window_name = "rlberry render"
        self.clipping_area = (-1.0, 1.0, -1.0, 1.0)

        # time counter
        self.time_count = 0

        # background scene
        self.background = Scene()
        # data to be rendered (list of scenes)
        self.data = []

    def set_window_name(self, name):
        self.window_name = name

    def set_refresh_interval(self, interval):
        self.refresh_interval = interval

    def set_clipping_area(self, area):
        """
        The clipping area is tuple with elements (left, right, bottom, top)
        Default = (-1.0, 1.0, -1.0, 1.0)
        """
        self.clipping_area = area
        base_size = max(self.window_width, self.window_height)
        width_range = area[1] - area[0]
        height_range = area[3] - area[2]
        base_range = max(width_range, height_range)
        width_range /= base_range
        height_range /= base_range
        self.window_width = int(base_size * width_range)
        self.window_height = int(base_size * height_range)

        # width and height must be divisible by 2
        if self.window_width % 2 == 1:
            self.window_width += 1
        if self.window_height % 2 == 1:
            self.window_height += 1

    def set_data(self, data):
        self.data = data

    def set_background(self, background):
        self.background = background

    def display(self):
        """
        Callback function, handler for window re-paint
        """
        # Set background color (clear background)
        self.screen.fill(self.background_color)

        # Display background
        for shape in self.background.shapes:
            self.draw_geometric2d(shape)

        # Display objects
        if len(self.data) > 0:
            idx = self.time_count % len(self.data)
            for shape in self.data[idx].shapes:
                self.draw_geometric2d(shape)

        self.time_count += 1

    def draw_geometric2d(self, shape):
        """
        Draw a 2D shape, of type GeometricPrimitive
        """
        if shape.type in ['POLYGON']:
            area = self.clipping_area
            width_range = area[1] - area[0]
            height_range = area[3] - area[2]

            vertices = []
            for vertex in shape.vertices:
                xx = vertex[0]*self.window_width/width_range
                yy = vertex[1]*self.window_height/height_range

                # put origin at bottom left instead of top left
                yy = self.window_height - yy

                pg_vertex = (xx, yy)
                vertices.append(pg_vertex)

            color = (255*shape.color[0],
                     255*shape.color[1],
                     255*shape.color[2])
            pg.draw.polygon(self.screen, color, vertices)

        else:
            raise NotImplementedError(
                "Shape type %s not implemented in pygame renderer."
                % shape.type)

    def run_graphics(self, debug_mode=False):
        """
        Sequentially displays scenes in self.data
        """
        global _IMPORT_SUCESSFUL

        if _IMPORT_SUCESSFUL:
            pg.init()
            display = (self.window_width, self.window_height)
            self.screen = pg.display.set_mode(display)
            pg.display.set_caption(self.window_name)
            while True:
                for event in pg.event.get():
                    if event.type == pg.QUIT:
                        pg.quit()
                        return
                #
                self.display()
                #
                pg.display.flip()
                pg.time.wait(self.refresh_interval)

                # in debug mode, stop
                if debug_mode:
                    pg.quit()
                    return
        else:
            print("Error: not possible to render the environment, \
pygame or pyopengl not installed.")

    def get_video_data(self):
        """
        Stores scenes in self.data in a list of numpy arrays that can be used
        to save a video.
        """
        global _IMPORT_SUCESSFUL

        if _IMPORT_SUCESSFUL:
            video_data = []

            pg.init()
            display = (self.window_width, self.window_height)
            self.screen = pg.display.set_mode(display)
            pg.display.set_caption(self.window_name)

            self.time_count = 0
            while self.time_count <= len(self.data):
                #
                self.display()
                #
                pg.display.flip()

                #
                # See https://stackoverflow.com/a/42754578/5691288
                #
                string_image = pg.image.tostring(self.screen, 'RGB')
                temp_surf = pg.image.fromstring(string_image,
                                                (self.window_width,
                                                 self.window_height), 'RGB')
                tmp_arr = pg.surfarray.array3d(temp_surf)
                imgdata = np.moveaxis(tmp_arr, 0, 1)
                video_data.append(imgdata)

            pg.quit()
            return video_data
        else:
            print("Error: not possible to render the environment, pygame \
or pyopengl not installed.")
            return []




from abc import ABC, abstractmethod
from copy import deepcopy


class Agent(ABC):
    """
    Basic interface for agents.

    Input environment is (deep)copied and reseeded.

    Notes
    ------
        Classes that implement this class should send **kwargs to
        Agent.__init__()


    Attributes
    ----------
    name : string
        agent identifier
    fit_info : tuple
        tuple of strings containing the keys in the dictionary
        returned by fit()
    env : rlberry.envs.interface.model.Model
        environment on which to train the agent


    Methods
    --------
    fit(**kwargs), abstract
        train the agent, returns dictionary with training info,
        whose keys are strings
    policy(observation, **kwargs), abstract
        returns the action to be taken given an observation
    reset()
        puts the agent in default setup (optional)
    sample_parameters(), optional
        returns dictionary with sampled hyperparameters.
    save(), optional
        save agent
    load(), optional
        load agent, returns an instance of the agent
    """

    name = ""
    fit_info = ()

    def __init__(self, env, copy_env=True,
                 reseed_env=True, **kwargs):
        """
        Parameters
        ----------
        env : Model
            Environment used to fit the agent.
        copy_env : bool
            If true, makes a deep copy of the environment.
        reseed_env : bool
            If true, reseeds the environment.
        """
        if copy_env:
            self.env = deepcopy(env)
        else:
            self.env = env
        if reseed_env:
            self.env.reseed()

    @abstractmethod
    def fit(self, **kwargs):
        """Train the agent using the provided environment."""
        pass

    @abstractmethod
    def policy(self, observation, **kwargs):
        """Returns an action, given an observation."""
        pass

    def reset(self, **kwargs):
        """Put the agent in default setup."""
        pass

    def save(self, **kwargs):
        """Save agent object."""
        raise NotImplementedError("agent.save() not implemented.")

    def load(self, **kwargs):
        """Load agent object."""
        raise NotImplementedError("agent.load() not implemented.")

    @classmethod
    def sample_parameters(cls, trial):
        """
        Sample hyperparameters for hyperparam optimization using
        Optuna (https://optuna.org/)

        Note: only the kwargs sent to __init__ are optimized. Make sure to
        include in the Agent constructor all "optimizable" parameters.

        Parameters
        ----------
        trial: optuna.trial
        """
        raise NotImplementedError("agent.sample_parameters() not implemented.")



import numpy as np

from rlberry.envs.finite import FiniteMDP
from rlberry.rendering import Scene, GeometricPrimitive, RenderInterface2D
from rlberry.rendering.common_shapes import circle_shape


class GridWorld(FiniteMDP, RenderInterface2D):
    """
    Simple GridWorld environment.

    Note: terminal states are not set to be absorbing in the transition array P
    """

    name = "GridWorld"

    def __init__(self,
                 nrows=5,
                 ncols=5,
                 start_coord=(0, 0),
                 terminal_states=None,
                 success_probability=0.9,
                 reward_at=None,
                 walls=((1, 1), (2, 2)),
                 default_reward=0.0):
        """
        Parameters
        -----------
        nrows : int
            number of rows
        ncols : int
            number of columns
        start_coord : tuple
            tuple with coordinates of initial position
        terminal_states : tuple
            ((row_0, col_0), (row_1, col_1), ...) = coordinates of
            terminal states
        success_probability : double
            probability of moving in the chosen direction
        reward_at: dict
            dictionary, keys = tuple containing coordinates, values = reward
            at each coordinate
        walls : tuple
            ((row_0, col_0), (row_1, col_1), ...) = coordinates of walls
        default_reward : double
            reward received at states not in  'reward_at'
        """
        # Grid dimensions
        self.nrows = nrows
        self.ncols = ncols

        # Reward parameters
        self.default_reward = default_reward

        # Default config
        if reward_at is not None:
            self.reward_at = reward_at
            if isinstance(next(iter(self.reward_at.keys())), str):
                self.reward_at = {eval(key): value for key, value
                                  in self.reward_at.items()}
        else:
            self.reward_at = {(nrows - 1, ncols - 1): 1}
        if walls is not None:
            self.walls = walls
        else:
            self.walls = ()
        if terminal_states is not None:
            self.terminal_states = terminal_states
        else:
            self.terminal_states = ((nrows - 1, ncols - 1),)

        # Probability of going left/right/up/down when choosing the
        # correspondent action
        # The remaining probability mass is distributed uniformly to other
        # available actions
        self.success_probability = success_probability

        # Start coordinate
        self.start_coord = tuple(start_coord)

        # Actions (string to index & index to string)
        self.a_str2idx = {'left': 0, 'right': 1, 'up': 2, 'down': 3}
        self.a_idx2str = {0: 'left', 1: 'right', 2: 'up', 3: 'down'}

        # --------------------------------------------
        # The variables below are defined in _build()
        # --------------------------------------------

        # Mappings (state index) <-> (state coordinate)
        self.index2coord = {}
        self.coord2index = {}

        # MDP parameters for base class
        self.P = None
        self.R = None
        self.Ns = None
        self.Na = 4

        # Build
        self._build()
        init_state_coord = self.coord2index[start_coord]
        FiniteMDP.__init__(self, self.R, self.P,
                           initial_state_distribution=init_state_coord)
        RenderInterface2D.__init__(self)
        self.reset()
        self.reward_range = (self.R.min(), self.R.max())

        # rendering info
        self.set_clipping_area((0, self.ncols, 0, self.nrows))
        self.set_refresh_interval(100)  # in milliseconds
        self.renderer_type = 'pygame'

    def is_terminal(self, state):
        state_coord = self.index2coord[state]
        return state_coord in self.terminal_states

    def reset(self, state=None):
        if state is None:
            state = self.coord2index[self.start_coord]
        self.state = state
        return state

    def reward_fn(self, state, action, next_state):
        row, col = self.index2coord[next_state]
        if (row, col) in self.reward_at:
            return self.reward_at[(row, col)]
        if (row, col) in self.walls:
            return 0.0
        return self.default_reward

    def _build(self):
        self._build_state_mappings_and_states()
        self._build_transition_probabilities()
        self._build_mean_rewards()

    def _build_state_mappings_and_states(self):
        index = 0
        for rr in range(self.nrows):
            for cc in range(self.ncols):
                if (rr, cc) in self.walls:
                    self.coord2index[(rr, cc)] = -1
                else:
                    self.coord2index[(rr, cc)] = index
                    self.index2coord[index] = (rr, cc)
                    index += 1
        states = np.arange(index).tolist()
        self.Ns = len(states)

    def _build_mean_rewards(self):
        S = self.Ns
        A = self.Na
        self.R = np.zeros((S, A))
        for ss in range(S):
            for aa in range(A):
                mean_r = 0
                for ns in range(S):
                    mean_r += self.reward_fn(ss, aa, ns) * self.P[ss, aa, ns]
                self.R[ss, aa] = mean_r

    def _build_transition_probabilities(self):
        Ns = self.Ns
        Na = self.Na
        self.P = np.zeros((Ns, Na, Ns))
        for s in range(Ns):
            s_coord = self.index2coord[s]
            neighbors = self._get_neighbors(*s_coord)
            valid_neighbors = [neighbors[nn][0] for nn in neighbors
                               if neighbors[nn][1]]
            n_valid = len(valid_neighbors)
            for a in range(Na):  # each action corresponds to a direction
                for nn in neighbors:
                    next_s_coord = neighbors[nn][0]
                    if next_s_coord in valid_neighbors:
                        next_s = self.coord2index[next_s_coord]
                        if a == nn:  # action is successful
                            self.P[s, a, next_s] = self.success_probability \
                                + (1 - self.success_probability) \
                                * (n_valid == 1)
                        elif neighbors[a][0] not in valid_neighbors:
                            self.P[s, a, s] = 1.0
                        else:
                            if n_valid > 1:
                                self.P[s, a, next_s] = \
                                    (1.0 - self.success_probability) \
                                    / (n_valid - 1)

    def _get_neighbors(self, row, col):
        aux = {}
        aux['left'] = (row, col - 1)  # left
        aux['right'] = (row, col + 1)  # right
        aux['up'] = (row - 1, col)  # up
        aux['down'] = (row + 1, col)  # down
        neighbors = {}
        for direction_str in aux:
            direction = self.a_str2idx[direction_str]
            next_s = aux[direction_str]
            neighbors[direction] = (next_s, self._is_valid(*next_s))
        return neighbors

    def get_transition_support(self, state):
        row, col = self.index2coord[state]
        neighbors = [(row, col - 1), (row, col + 1),
                     (row - 1, col), (row + 1, col)]
        return [self.coord2index[coord] for coord in neighbors
                if self._is_valid(*coord)]

    def _is_valid(self, row, col):
        if (row, col) in self.walls:
            return False
        elif row < 0 or row >= self.nrows:
            return False
        elif col < 0 or col >= self.ncols:
            return False
        return True

    def _build_ascii(self):
        grid = [[''] * self.ncols for rr in range(self.nrows)]
        grid_idx = [[''] * self.ncols for rr in range(self.nrows)]
        for rr in range(self.nrows):
            for cc in range(self.ncols):
                if (rr, cc) in self.walls:
                    grid[rr][cc] = 'x '
                else:
                    grid[rr][cc] = 'o '
                grid_idx[rr][cc] = str(self.coord2index[(rr, cc)]).zfill(3)

        for (rr, cc) in self.reward_at:
            rwd = self.reward_at[(rr, cc)]
            if rwd > 0:
                grid[rr][cc] = '+ '
            if rwd < 0:
                grid[rr][cc] = '-'

        grid[self.start_coord[0]][self.start_coord[1]] = 'I '

        # current position of the agent
        x, y = self.index2coord[self.state]
        grid[x][y] = 'A '

        #
        grid_ascii = ''
        for rr in range(self.nrows + 1):
            if rr < self.nrows:
                grid_ascii += str(rr).zfill(2) + 2 * ' '\
                     + ' '.join(grid[rr]) + '\n'
            else:
                grid_ascii += 3 * ' ' + ' '.join([str(jj).zfill(2) for jj
                                                  in range(self.ncols)])

        self.grid_ascii = grid_ascii
        self.grid_idx = grid_idx
        return self.grid_ascii

    def display_values(self, values):
        assert len(values) == self.Ns
        grid_values = [['X'.ljust(9)] * self.ncols for ii in range(self.nrows)]
        for s_idx in range(self.Ns):
            v = values[s_idx]
            row, col = self.index2coord[s_idx]
            grid_values[row][col] = ("%0.2f" % v).ljust(9)

        grid_values_ascii = ''
        for rr in range(self.nrows + 1):
            if rr < self.nrows:
                grid_values_ascii += str(rr).zfill(2) + 2 * ' ' \
                    + ' '.join(grid_values[rr]) + '\n'
            else:
                grid_values_ascii += 4 * ' ' \
                    + ' '.join([str(jj).zfill(2).ljust(9) for jj
                                in range(self.ncols)])
        print(grid_values_ascii)

    def print_transition_at(self, row, col, action):
        s_idx = self.coord2index[(row, col)]
        if s_idx < 0:
            print("wall!")
            return
        a_idx = self.a_str2idx[action]
        for next_s_idx, prob in enumerate(self.P[s_idx, a_idx]):
            if prob > 0:
                print("to (%d, %d) with prob %f" %
                      (self.index2coord[next_s_idx] + (prob,)))

    def render_ascii(self):
        print(self._build_ascii())

    def step(self, action):
        assert self.action_space.contains(action), "Invalid action!"

        # save state for rendering
        if self.is_render_enabled():
            self.append_state_for_rendering(self.state)

        # take step
        next_state, reward, done, info = self.sample(self.state, action)
        self.state = next_state
        return next_state, reward, done, info

    #
    # Code for rendering
    #

    def get_background(self):
        """
        Returne a scene (list of shapes) representing the background
        """
        bg = Scene()

        # walls
        for wall in self.walls:
            y, x = wall
            shape = GeometricPrimitive("POLYGON")
            shape.set_color((0.25, 0.25, 0.25))
            shape.add_vertex((x, y))
            shape.add_vertex((x + 1, y))
            shape.add_vertex((x + 1, y + 1))
            shape.add_vertex((x, y + 1))
            bg.add_shape(shape)

        # rewards
        for (y, x) in self.reward_at:
            flag = GeometricPrimitive("POLYGON")
            rwd = self.reward_at[(y, x)]
            if rwd > 0:
                flag.set_color((0.0, 0.5, 0.0))
            if rwd < 0:
                flag.set_color((0.5, 0.0, 0.0))

            x += 0.5
            y += 0.25
            flag.add_vertex((x, y))
            flag.add_vertex((x + 0.25, y + 0.5))
            flag.add_vertex((x - 0.25, y + 0.5))
            bg.add_shape(flag)

        return bg

    def get_scene(self, state):
        """
        Return scene (list of shapes) representing a given state
        """
        y, x = self.index2coord[state]
        x = x + 0.5  # centering
        y = y + 0.5  # centering

        scene = Scene()

        agent = circle_shape((x, y), 0.25, n_points=5)
        agent.type = "POLYGON"
        agent.set_color((0.75, 0.0, 0.5))

        scene.add_shape(agent)
        return scene


# if __name__ == '__main__':
#     env = GridWorld(nrows=5, ncols=5,
#                     reward_at={(4, 4): 1, (4, 3): -1})
#     env.step(env.action_space.sample())
#     env.render_ascii()



import numpy as np
from rlberry.wrappers import Wrapper


class RescaleRewardWrapper(Wrapper):
    """
    Rescale the reward function to a bounded range.
    """

    def __init__(self, env, reward_range):
        """
        Parameters
        ----------
        reward_range: tuple (double, double)
            tuple with the desired reward range, which needs to be bounded.
        """
        Wrapper.__init__(self, env)
        self.reward_range = reward_range
        assert reward_range[0] < reward_range[1]
        assert reward_range[0] > -np.inf and reward_range[1] < np.inf

    def _linear_rescaling(self, x, x0, x1, u0, u1):
        """
        For x a value in [x0, x1], maps x linearly to the interval [u0, u1].
        """
        a = (u1 - u0) / (x1 - x0)
        b = (x1 * u0 - x0 * u1) / (x1 - x0)
        return a * x + b

    def _rescale(self, reward):
        x0, x1 = self.env.reward_range
        u0, u1 = self.reward_range
        # bounded reward
        if x0 > -np.inf and x1 < np.inf:
            return self._linear_rescaling(reward, x0, x1, u0, u1)
        # unbounded
        elif x0 > -np.inf and x1 == np.inf:
            x = reward - x0  # [0, infty]
            x = 2.0 / (1.0 + np.exp(-x)) - 1.0  # [0, 1]
            return self._linear_rescaling(x, 0.0, 1.0, u0, u1)
            # unbouded below
        elif x0 == -np.inf and x1 < np.inf:
            x = reward - x1  # [-infty, 0]
            x = 2.0 / (1.0 + np.exp(-x))  # [0, 1]
            return self._linear_rescaling(x, 0.0, 1.0, u0, u1)
            # unbounded
        else:
            x = 1.0 / (1.0 + np.exp(-reward))  # [0, 1]
            return self._linear_rescaling(x, 0.0, 1.0, u0, u1)

    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        rescaled_reward = self._rescale(reward)
        return observation, rescaled_reward, done, info

    def sample(self, state, action):
        observation, reward, done, info = self.env.sample(state, action)
        rescaled_reward = self._rescale(reward)
        return observation, rescaled_reward, done, info



"""
Mountain Car environment adapted from OpenAI gym [1].

* default reward is 0       (instead of -1)
* reward in goal state is 1 (instead of 0)
* also implemented as a generative model (in addition to an online model)
* render function follows the rlberry rendering interface.

[1] https://github.com/openai/gym/blob/master/gym/envs/
classic_control/mountain_car.py
"""

import math

import numpy as np

import rlberry.spaces as spaces
from rlberry.envs.interface import Model
from rlberry.rendering import Scene, GeometricPrimitive, RenderInterface2D


class MountainCar(Model, RenderInterface2D):
    """
    Description:
        The agent (a car) is started at the bottom of a valley. For any given
        state the agent may choose to accelerate to the left, right or cease
        any acceleration.

    Source:
        The environment appeared first in Andrew Moore's PhD Thesis (1990).

    Observation:
        Type: Box(2)
        Num    Observation               Min            Max
        0      Car Position              -1.2           0.6
        1      Car Velocity              -0.07          0.07

    Actions:
        Type: Discrete(3)
        Num    Action
        0      Accelerate to the Left
        1      Don't accelerate
        2      Accelerate to the Right

        Note: This does not affect the amount of velocity affected by the
        gravitational pull acting on the car.

    Reward:
         Reward of 1 is awarded if the agent reached the flag (position = 0.5)
         on top of the mountain.
         Reward of 0 is awarded if the position of the agent is less than 0.5.

    Starting State:
         The position of the car is assigned a uniform random value in
         [-0.6 , -0.4].
         The starting velocity of the car is always assigned to 0.

     Episode Termination:
         The car position is more than 0.5
     """

    name = "MountainCar"

    def __init__(self, goal_velocity=0):
        # init base classes
        Model.__init__(self)
        RenderInterface2D.__init__(self)

        self.min_position = -1.2
        self.max_position = 0.6
        self.max_speed = 0.07
        self.goal_position = 0.5
        self.goal_velocity = goal_velocity

        self.force = 0.001
        self.gravity = 0.0025

        self.low = np.array([self.min_position, -self.max_speed])
        self.high = np.array([self.max_position, self.max_speed])

        self.action_space = spaces.Discrete(3)
        self.observation_space = spaces.Box(self.low, self.high)

        self.reward_range = (0.0, 1.0)

        # rendering info
        self.set_clipping_area((-1.2, 0.6, -0.2, 1.1))
        self.set_refresh_interval(10)  # in milliseconds

        # initial reset
        self.reset()

    def step(self, action):
        assert self.action_space.contains(action), \
            "%r (%s) invalid" % (action,  type(action))

        # save state for rendering
        if self.is_render_enabled():
            self.append_state_for_rendering(np.array(self.state))

        next_state, reward, done, info = self.sample(self.state, action)
        self.state = next_state.copy()

        return next_state, reward, done, info

    def reset(self):
        self.state = np.array([self.rng.uniform(low=-0.6, high=-0.4), 0])
        return self.state.copy()

    def sample(self, state, action):
        if (type(state) != np.ndarray):
            state = np.array(state)
        assert self.observation_space.contains(state), \
            "Invalid state as argument of reset()."
        assert self.action_space.contains(action), \
            "%r (%s) invalid" % (action, type(action))

        position = state[0]
        velocity = state[1]
        velocity += (action - 1) * self.force \
            + math.cos(3 * position) * (-self.gravity)
        velocity = np.clip(velocity, -self.max_speed, self.max_speed)
        position += velocity
        position = np.clip(position, self.min_position, self.max_position)
        if (position == self.min_position and velocity < 0):
            velocity = 0

        done = bool(position >= self.goal_position and
                    velocity >= self.goal_velocity)
        reward = 0.0
        if done:
            reward = 1.0

        next_state = np.array([position, velocity])
        return next_state, reward, done, {}

    def _height(self, xs):
        return np.sin(3 * xs) * .45 + .55

    #
    # Below: code for rendering
    #

    def get_background(self):
        bg = Scene()
        mountain = GeometricPrimitive("TRIANGLE_FAN")
        flag = GeometricPrimitive("TRIANGLES")
        mountain.set_color((0.6, 0.3, 0.0))
        flag.set_color((0.0, 0.5, 0.0))

        # Mountain
        mountain.add_vertex((-0.3, -1.0))
        mountain.add_vertex((0.6, -1.0))

        n_points = 50
        obs_range = self.observation_space.high[0] \
            - self.observation_space.low[0]
        eps = obs_range / (n_points - 1)
        for ii in reversed(range(n_points)):
            x = self.observation_space.low[0] + ii * eps
            y = self._height(x)
            mountain.add_vertex((x, y))
        mountain.add_vertex((-1.2, -1.0))

        # Flag
        goal_x = self.goal_position
        goal_y = self._height(goal_x)
        flag.add_vertex((goal_x, goal_y))
        flag.add_vertex((goal_x + 0.025, goal_y + 0.075))
        flag.add_vertex((goal_x - 0.025, goal_y + 0.075))

        bg.add_shape(mountain)
        bg.add_shape(flag)

        return bg

    def get_scene(self, state):
        scene = Scene()

        agent = GeometricPrimitive("QUADS")
        agent.set_color((0.0, 0.0, 0.0))
        size = 0.025
        x = state[0]
        y = self._height(x)
        agent.add_vertex((x - size, y - size))
        agent.add_vertex((x + size, y - size))
        agent.add_vertex((x + size, y + size))
        agent.add_vertex((x - size, y + size))

        scene.add_shape(agent)
        return scene



import pytest
import numpy as np
from rlberry.utils.metrics import metric_lp


@pytest.mark.parametrize("dim", [1, 2, 3, 4])
def test_metrics(dim):
    y = np.zeros(dim)
    x = np.ones(dim)
    scaling_1 = np.ones(dim)
    scaling_2 = 0.5*np.ones(dim)

    for p in range(1, 10):
        assert np.abs(metric_lp(x, y, p, scaling_1)
                      - np.power(dim, 1.0/p)) < 1e-15
        assert np.abs(metric_lp(x, y, p, scaling_2)
                      - 2*np.power(dim, 1.0/p)) < 1e-15



"""
OpenGL code for 2D rendering, using pygame.
"""

import numpy as np
from os import environ
from rlberry.rendering import Scene

environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'

_IMPORT_SUCESSFUL = True
try:
    import pygame as pg
    from pygame.locals import DOUBLEBUF, OPENGL

    from OpenGL.GLU import gluOrtho2D
    from OpenGL.GL import glMatrixMode, glLoadIdentity, glClearColor
    from OpenGL.GL import glClear, glFlush, glBegin, glEnd
    from OpenGL.GL import glColor3f, glVertex2f
    from OpenGL.GL import GL_PROJECTION, GL_COLOR_BUFFER_BIT
    from OpenGL.GL import GL_POINTS, GL_LINES, GL_LINE_STRIP, GL_LINE_LOOP
    from OpenGL.GL import GL_POLYGON, GL_TRIANGLES, GL_TRIANGLE_STRIP
    from OpenGL.GL import GL_TRIANGLE_FAN, GL_QUADS, GL_QUAD_STRIP
except Exception:
    _IMPORT_SUCESSFUL = False


class OpenGLRender2D:
    """
    Class to render a list of scenes using OpenGL and pygame.
    """

    def __init__(self):
        # parameters
        self.window_width = 800
        self.window_height = 800    # multiples of 16 are preferred
        self.background_color = (0.6, 0.75, 1.0)
        self.refresh_interval = 50
        self.window_name = "rlberry render"
        self.clipping_area = (-1.0, 1.0, -1.0, 1.0)

        # time counter
        self.time_count = 0

        # background scene
        self.background = Scene()
        # data to be rendered (list of scenes)
        self.data = []

    def set_window_name(self, name):
        self.window_name = name

    def set_refresh_interval(self, interval):
        self.refresh_interval = interval

    def set_clipping_area(self, area):
        """
        The clipping area is tuple with elements (left, right, bottom, top)
        Default = (-1.0, 1.0, -1.0, 1.0)
        """
        self.clipping_area = area
        base_size = max(self.window_width, self.window_height)
        width_range = area[1] - area[0]
        height_range = area[3] - area[2]
        base_range = max(width_range, height_range)
        width_range /= base_range
        height_range /= base_range
        self.window_width = int(base_size * width_range)
        self.window_height = int(base_size * height_range)

        # width and height must be divisible by 2
        if self.window_width % 2 == 1:
            self.window_width += 1
        if self.window_height % 2 == 1:
            self.window_height += 1

    def set_data(self, data):
        self.data = data

    def set_background(self, background):
        self.background = background

    def initGL(self):
        """
        initialize GL
        """
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluOrtho2D(self.clipping_area[0], self.clipping_area[1],
                   self.clipping_area[2], self.clipping_area[3])

    def display(self):
        """
        Callback function, handler for window re-paint
        """
        # Set background color (clear background)
        glClearColor(self.background_color[0], self.background_color[1],
                     self.background_color[2], 1.0)
        glClear(GL_COLOR_BUFFER_BIT)

        # Display background
        for shape in self.background.shapes:
            self.draw_geometric2d(shape)

        # Display objects
        if len(self.data) > 0:
            idx = self.time_count % len(self.data)
            for shape in self.data[idx].shapes:
                self.draw_geometric2d(shape)

        self.time_count += 1
        glFlush()

    def draw_geometric2d(self, shape):
        """
        Draw a 2D shape, of type GeometricPrimitive
        """
        if shape.type == "POINTS":
            glBegin(GL_POINTS)
        elif shape.type == "LINES":
            glBegin(GL_LINES)
        elif shape.type == "LINE_STRIP":
            glBegin(GL_LINE_STRIP)
        elif shape.type == "LINE_LOOP":
            glBegin(GL_LINE_LOOP)
        elif shape.type == "POLYGON":
            glBegin(GL_POLYGON)
        elif shape.type == "TRIANGLES":
            glBegin(GL_TRIANGLES)
        elif shape.type == "TRIANGLE_STRIP":
            glBegin(GL_TRIANGLE_STRIP)
        elif shape.type == "TRIANGLE_FAN":
            glBegin(GL_TRIANGLE_FAN)
        elif shape.type == "QUADS":
            glBegin(GL_QUADS)
        elif shape.type == "QUAD_STRIP":
            glBegin(GL_QUAD_STRIP)
        else:
            print("Invalid type for geometric primitive!")
            raise NameError

        # set color
        glColor3f(shape.color[0], shape.color[1], shape.color[2])

        # create vertices
        for vertex in shape.vertices:
            glVertex2f(vertex[0], vertex[1])
        glEnd()

    def run_graphics(self, debug_mode=False):
        """
        Sequentially displays scenes in self.data
        """
        global _IMPORT_SUCESSFUL

        if _IMPORT_SUCESSFUL:
            pg.init()
            display = (self.window_width, self.window_height)
            pg.display.set_mode(display, DOUBLEBUF | OPENGL)
            pg.display.set_caption(self.window_name)
            self.initGL()
            while True:
                for event in pg.event.get():
                    if event.type == pg.QUIT:
                        pg.quit()
                        return
                #
                self.display()
                #
                pg.display.flip()
                pg.time.wait(self.refresh_interval)

                # in debug mode, stop
                if debug_mode:
                    pg.quit()
                    return
        else:
            print("Error: not possible to render the environment, \
pygame or pyopengl not installed.")

    def get_video_data(self):
        """
        Stores scenes in self.data in a list of numpy arrays that can be used
        to save a video.
        """
        global _IMPORT_SUCESSFUL

        if _IMPORT_SUCESSFUL:
            video_data = []

            pg.init()
            display = (self.window_width, self.window_height)
            screen = pg.display.set_mode(display, DOUBLEBUF | OPENGL)
            pg.display.set_caption(self.window_name)
            self.initGL()

            self.time_count = 0
            while self.time_count <= len(self.data):
                #
                self.display()
                #
                pg.display.flip()

                #
                # See https://stackoverflow.com/a/42754578/5691288
                #
                string_image = pg.image.tostring(screen, 'RGB')
                temp_surf = pg.image.fromstring(string_image,
                                                (self.window_width,
                                                 self.window_height), 'RGB')
                tmp_arr = pg.surfarray.array3d(temp_surf)
                imgdata = np.moveaxis(tmp_arr, 0, 1)
                video_data.append(imgdata)

            pg.quit()
            return video_data
        else:
            print("Error: not possible to render the environment, \
pygame or pyopengl not installed.")
            return []



import numpy as np
import logging

_FFMPEG_INSTALLED = True
try:
    import ffmpeg
except Exception:
    _FFMPEG_INSTALLED = False


def video_write(fn, images, framerate=60, vcodec='libx264'):
    """
    Save list of images to a video file.

    Source:
    https://github.com/kkroening/ffmpeg-python/issues/246#issuecomment-520200981
    Modified so that framerate is given to .input(), as suggested in the
    thread, to avoid
    skipping frames.

    Parameters
    ----------
    fn : string
        filename
    images : list or np.array
        list of images to save to a video.
    framerate : int
    """
    global _FFMPEG_INSTALLED

    try:
        if len(images) == 0:
            logging.warning("Calling video_write() with empty images.")
            return

        if not _FFMPEG_INSTALLED:
            logging.error(
                "video_write(): Unable to save video, ffmpeg-python \
    package required (https://github.com/kkroening/ffmpeg-python)")
            return

        if not isinstance(images, np.ndarray):
            images = np.asarray(images)
        n, height, width, channels = images.shape
        process = (
            ffmpeg
            .input('pipe:', format='rawvideo', pix_fmt='rgb24',
                s='{}x{}'.format(width, height), r=framerate)
            .output(fn, pix_fmt='yuv420p', vcodec=vcodec)
            .overwrite_output()
            .run_async(pipe_stdin=True)
        )
        for frame in images:
            process.stdin.write(
                frame
                .astype(np.uint8)
                .tobytes()
            )
        process.stdin.close()
        process.wait()

    except Exception as ex:
        logging.warning("Not possible to save \
video, due to exception: {}".format(str(ex)))



import numpy as np
import pytest
import rlberry.seeding as seeding

from rlberry.envs.classic_control import MountainCar
from rlberry.envs.finite import FiniteMDP
from rlberry.wrappers.discretize_state import DiscretizeStateWrapper
from rlberry.wrappers.rescale_reward import RescaleRewardWrapper
from rlberry.wrappers.autoreset import AutoResetWrapper


@pytest.mark.parametrize("n_bins", list(range(1, 10)))
def test_discretizer(n_bins):
    env = DiscretizeStateWrapper(MountainCar(), n_bins)
    assert env.observation_space.n == n_bins * n_bins

    for _ in range(2):
        state = env.reset()
        for _ in range(50):
            assert env.observation_space.contains(state)
            action = env.action_space.sample()
            next_s, _, _, _ = env.step(action)
            state = next_s

    for _ in range(100):
        state = env.observation_space.sample()
        action = env.action_space.sample()
        next_s, _, _, _ = env.sample(state, action)
        assert env.observation_space.contains(next_s)

    assert env.unwrapped.name == "MountainCar"


def test_rescale_reward():
    # tolerance
    tol = 1e-14

    rng = seeding.get_rng()

    for _ in range(10):
        # generate random MDP
        S, A = 5, 2
        R = rng.uniform(0.0, 1.0, (S, A))
        P = rng.uniform(0.0, 1.0, (S, A, S))
        for ss in range(S):
            for aa in range(A):
                P[ss, aa, :] /= P[ss, aa, :].sum()
        env = FiniteMDP(R, P)

        # test
        wrapped = RescaleRewardWrapper(env, (-10, 10))
        _ = wrapped.reset()
        for _ in range(100):
            _, reward, _, _ = wrapped.sample(
                wrapped.observation_space.sample(),
                wrapped.action_space.sample())
            assert reward <= 10+tol and reward >= -10-tol

        _ = wrapped.reset()
        for _ in range(100):
            _, reward, _, _ = wrapped.step(wrapped.action_space.sample())
            assert reward <= 10+tol and reward >= -10-tol


@pytest.mark.parametrize("rmin, rmax", [(0, 1), (-1, 1), (-5, 5), (-5, 15)])
def test_rescale_reward_2(rmin, rmax):
    # tolerance
    tol = 1e-15

    # dummy MDP
    S, A = 5, 2
    R = np.ones((S, A))
    P = np.ones((S, A, S))
    for ss in range(S):
        for aa in range(A):
            P[ss, aa, :] /= P[ss, aa, :].sum()
    env = FiniteMDP(R, P)

    # test bounded case
    env.reward_range = (-100, 50)
    wrapped = RescaleRewardWrapper(env, (rmin, rmax))
    xx = np.linspace(-100, 50, num=100)
    for x in xx:
        y = wrapped._rescale(x)
        assert y >= rmin-tol and y <= rmax+tol

    # test unbounded above
    env.reward_range = (-1.0, np.inf)
    wrapped = RescaleRewardWrapper(env, (rmin, rmax))
    xx = np.linspace(-1, 1e2, num=100)
    for x in xx:
        y = wrapped._rescale(x)
        assert y >= rmin-tol and y <= rmax+tol

    # test unbounded below
    env.reward_range = (-np.inf, 1.0)
    wrapped = RescaleRewardWrapper(env, (rmin, rmax))
    xx = np.linspace(-1e2, 1, num=100)
    for x in xx:
        y = wrapped._rescale(x)
        assert y >= rmin-tol and y <= rmax+tol

    # test unbounded
    env.reward_range = (-np.inf, np.inf)
    wrapped = RescaleRewardWrapper(env, (rmin, rmax))
    xx = np.linspace(-1e2, 1e2, num=200)
    for x in xx:
        y = wrapped._rescale(x)
        assert y >= rmin-tol and y <= rmax+tol


@pytest.mark.parametrize("horizon", list(range(1, 10)))
def test_autoreset(horizon):

    # dummy MDP
    S, A = 5, 2
    R = np.ones((S, A))
    P = np.ones((S, A, S))
    for ss in range(S):
        for aa in range(A):
            P[ss, aa, :] /= P[ss, aa, :].sum()
    # initial state = 3
    env = FiniteMDP(R, P, initial_state_distribution=3)
    env = AutoResetWrapper(env, horizon)

    env.reset()
    for tt in range(5*horizon+1):
        action = env.action_space.sample()
        next_s, reward, done, info = env.step(action)
        if (tt+1) % horizon == 0:
            assert next_s == 3



import numpy as np
from rlberry.utils.jit_setup import numba_jit
from rlberry.utils.metrics import metric_lp


@numba_jit
def map_to_representative(state,
                          lp_metric,
                          representative_states,
                          n_representatives,
                          min_dist,
                          scaling,
                          accept_new_repr):
    """Map state to representative state. """
    dist_to_closest = np.inf
    argmin = -1
    for ii in range(n_representatives):
        dist = metric_lp(state, representative_states[ii, :],
                         lp_metric,
                         scaling)
        if dist < dist_to_closest:
            dist_to_closest = dist
            argmin = ii

    max_representatives = representative_states.shape[0]
    if (dist_to_closest > min_dist) \
        and (n_representatives < max_representatives) \
            and accept_new_repr:
        new_index = n_representatives
        representative_states[new_index, :] = state
        return new_index
    return argmin



from abc import abstractmethod
from rlberry.agents import Agent


class IncrementalAgent(Agent):
    """Basic interface for agents that can be trained incrementally."""

    name = ""
    fit_info = ()

    def __init__(self, env, **kwargs):
        """

        Parameters
        ----------
        env : Model
            Environment used to fit the agent.
        """
        Agent.__init__(self, env, **kwargs)

    @abstractmethod
    def partial_fit(self, fraction, **kwargs):
        """Partially fits the agent, according to the fraction parameter.

        For instance, if the agent requires N episodes for a "full" fit,
        calling partial_fit(0.5) will fit the agent for 0.5*N episodes.

        Also, calling partial_fit(0.5) twice must be equivalent to
        a single call to fit().

        Parameters
        ---------
        fraction: double, in [0,1]
            Fraction of the agent to fit.

        Returns
        -------
        info : dict
        """
        raise NotImplementedError("agent.partial_fit() not implemented.")




import time

import numpy as np
import torch
import torch.nn as nn

import rlberry.seeding as seeding
import rlberry.spaces as spaces
from rlberry.agents import Agent

# choose device
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class Net(nn.Module):
    def __init__(self, obs_size, hidden_size, n_actions):
        super(Net, self).__init__()
        self.net = nn.Sequential(
            nn.Linear(obs_size, hidden_size),
            nn.ReLU(),
            nn.Linear(hidden_size, n_actions)
        )

    def forward(self, x):
        return self.net(x)


class Memory:
    def __init__(self, max_size):
        self.max_size = max_size
        self.clear()

    def clear(self):
        self.size = 0
        self.states = []
        self.actions = []
        self.rewards = []

    def append(self, state, action, reward):
        self.states.append(state)
        self.actions.append(action)
        self.rewards.append(reward)
        self.size += 1
        if self.size == self.max_size+1:
            self.states.pop(0)
            self.actions.pop(0)
            self.rewards.pop(0)
            self.size = self.max_size


class CEMAgent(Agent):

    name = "CrossEntropyAgent"
    fit_info = ("n_episodes", "episode_rewards")

    def __init__(self,
                 env,
                 n_episodes=1000,
                 horizon=100,
                 gamma=0.99,
                 batch_size=16,
                 percentile=70,
                 learning_rate=0.01,
                 verbose=5,
                 **kwargs):
        """
        Parameters
        ----------
        env : Model
            Environment for training.
        n_episodes : int
            Number of training episodes.
        horizon : int
            Maximum length of a trajectory.
        gamma : double
            Discount factor in [0, 1].
        batch_size : int
            Number of trajectories to sample at each iteration.
        percentile : int
            Percentile used to remove trajectories with low rewards.
        learning_rate : double
            Optimizer learning rate
        verbose : int
            Verbosity level.
        """
        Agent.__init__(self, env, **kwargs)

        # check environment
        assert isinstance(self.env.observation_space, spaces.Box)
        assert isinstance(self.env.action_space, spaces.Discrete)

        # parameters
        self.gamma = gamma
        self.batch_size = batch_size
        self.n_episodes = n_episodes
        self.percentile = percentile
        self.learning_rate = learning_rate
        self.horizon = horizon
        self.verbose = verbose

        # random number generator
        self.rng = seeding.get_rng()

        # policy net
        hidden_size = 128
        obs_size = self.env.observation_space.high.shape[0]
        n_actions = self.env.action_space.n
        self.policy_net = Net(obs_size, hidden_size, n_actions).to(device)

        # loss function and optimizer
        self.loss_fn = nn.CrossEntropyLoss()
        self.optimizer = torch.optim.Adam(params=self.policy_net.parameters(),
                                          lr=self.learning_rate)

        # memory
        self.memory = Memory(self.batch_size)

        # logging config
        self._last_printed_ep = 0
        self._time_last_log = time.process_time()
        if self.verbose == 1:
            self._log_interval = 60  # in seconds
        elif self.verbose == 2:
            self._log_interval = 30
        elif self.verbose == 3:
            self._log_interval = 15
        elif self.verbose > 3:
            self._log_interval = 5

    def fit(self, **kwargs):
        self.episode = 0
        info = {}
        self._rewards = np.zeros(self.n_episodes)
        self._cumul_rewards = np.zeros(self.n_episodes)

        for ep in range(self.n_episodes):
            episode_rewards = self._run_episode()
            self._rewards[ep] = episode_rewards
            self._cumul_rewards[ep] = episode_rewards + \
                self._cumul_rewards[max(0, ep - 1)]

            # update policy and clear batch
            loss, reward_mean, reward_bound = self._update()
            self.episode += 1
            self._logging()

        info["n_episodes"] = self.n_episodes
        info["episode_rewards"] = self._rewards
        return info

    def _logging(self):
        if self.verbose > 0:
            t_now = time.process_time()
            time_elapsed = t_now - self._time_last_log
            if (time_elapsed >= self._log_interval) \
                    or (self.episode == self.n_episodes):
                self._time_last_log = t_now
                print(self._info_to_print())
                self._last_printed_ep = self.episode - 1

    def _info_to_print(self):
        prev_episode = self._last_printed_ep
        episode = self.episode - 1
        reward_per_ep = self._rewards[prev_episode:episode + 1].sum() \
            / max(1, episode - prev_episode)
        time_per_ep = self._log_interval * 1000.0 \
            / max(1, episode - prev_episode)
        time_per_ep = max(0.01, time_per_ep)  # avoid div by zero
        fps = int((self.horizon / time_per_ep) * 1000)

        to_print = "[{}] episode = {}/{} ".format(self.name, episode+1,
                                                  self.n_episodes) \
            + "| reward/ep = {:0.2f} ".format(reward_per_ep) \
            + "| time/ep = {:0.2f} ms".format(time_per_ep) \
            + "| fps = {}".format(fps)
        return to_print

    def policy(self, observation, **kwargs):
        act_probs_v, scores = \
            self._get_action_probabilities_tensor(observation)
        act_probs = act_probs_v.data.numpy()[0]
        action = self.rng.choice(len(act_probs), p=act_probs)
        return action

    def _get_action_probabilities_tensor(self, observation):
        obs_v = torch.FloatTensor([observation]).to(device)
        sm = nn.Softmax(dim=1)
        scores = self.policy_net(obs_v)
        act_probs_v = sm(scores)
        return act_probs_v, scores

    def _process_batch(self):
        rewards = np.array(self.memory.rewards)

        reward_bound = np.percentile(rewards, self.percentile)
        reward_mean = float(np.mean(rewards))

        train_states = []
        train_actions = []

        for ii in range(self.memory.size):
            if rewards[ii] < reward_bound:
                continue
            train_states.extend(self.memory.states[ii])
            train_actions.extend(self.memory.actions[ii])

        train_states_tensor = torch.FloatTensor(train_states).to(device)
        train_actions_tensor = torch.LongTensor(train_actions).to(device)

        # states in last trajectory
        last_states = self.memory.states[-1]
        last_states_tensor = torch.FloatTensor(last_states).to(device)

        return train_states_tensor, train_actions_tensor, \
            reward_bound, reward_mean, last_states_tensor

    def _run_episode(self):
        # interact for H steps
        episode_rewards = 0
        episode_states = []
        episode_actions = []
        state = self.env.reset()
        for hh in range(self.horizon):
            # take action according to policy_net
            action = self.policy(state)
            next_state, reward, done, _ = self.env.step(action)

            # save
            episode_states.append(state)
            episode_actions.append(action)

            # increment rewards
            episode_rewards += reward*np.power(self.gamma, hh)

            if done:
                break
            state = next_state

        self.memory.append(episode_states, episode_actions, episode_rewards)

        return episode_rewards

    def _update(self):
        train_states_tensor, train_actions_tensor,\
            reward_bound, reward_mean, \
            last_states_tensor = self._process_batch()
        self.optimizer.zero_grad()
        action_scores = self.policy_net(train_states_tensor)
        loss = self.loss_fn(action_scores, train_actions_tensor)

        # entropy in last trajectory
        scores_last_traj = self.policy_net(last_states_tensor)
        softmax = nn.Softmax(dim=1)
        # 1e-3 is added to avoid zeros
        probs_last_traj = softmax(scores_last_traj) + 1e-3
        entropy = - torch.sum((probs_last_traj * torch.log(probs_last_traj)),
                              dim=1)
        entropy = torch.mean(entropy)

        loss = loss - 0.1*entropy
        loss.backward()
        self.optimizer.step()

        return loss.item(), reward_mean, reward_bound



import numpy as np


def binary_search_nd(x_vec, bins):
    """n-dimensional binary search

    Parameters
    -----------
    x_vec : numpy.ndarray
        numpy 1d array to be searched in the bins
    bins : list
        list of numpy 1d array, bins[d] = bins of the d-th dimension


    Returns
    --------
    index (int) corresponding to the position of x in the partition
    defined by the bins.
    """
    dim = len(bins)
    flat_index = 0
    aux = 1
    assert dim == len(x_vec), "dimension mismatch in binary_search_nd()"
    for dd in range(dim):
        index_dd = np.searchsorted(bins[dd], x_vec[dd], side='right') - 1
        assert index_dd != -1, "error in binary_search_nd()"
        flat_index += aux * index_dd
        aux *= (len(bins[dd]) - 1)
    return flat_index


def unravel_index_uniform_bin(flat_index, dim, n_per_dim):
    index = []
    aux_index = flat_index
    for _ in range(dim):
        index.append(aux_index % n_per_dim)
        aux_index = aux_index // n_per_dim
    return tuple(index)


if __name__ == "__main__":
    bins = [(0, 1, 2, 3, 4),
            (0, 1, 2, 3, 4)]
    x = [3.9, 3.5]
    index = binary_search_nd(x, bins)
    print(index)




import numpy as np
from rlberry.exploration_tools.uncertainty_estimator \
    import UncertaintyEstimator
from rlberry.spaces import Discrete
from rlberry.utils.space_discretizer import Discretizer


class DiscreteCounter(UncertaintyEstimator):
    def __init__(self, observation_space, action_space, n_bins_obs=10,
                 n_bins_actions=10, **kwargs):
        UncertaintyEstimator.__init__(self, observation_space, action_space)

        self.continuous_state = False
        self.continuous_action = False

        if isinstance(observation_space, Discrete):
            self.n_states = observation_space.n
        else:
            self.continuous_state = True
            self.state_discretizer = Discretizer(self.observation_space,
                                                 n_bins_obs)
            self.n_states = self.state_discretizer.discrete_space.n

        if isinstance(action_space, Discrete):
            self.n_actions = action_space.n
        else:
            self.continuous_action = True
            self.action_discretizer = Discretizer(self.action_space,
                                                  n_bins_actions)
            self.n_actions = self.action_discretizer.discrete_space.n

        self.N_sa = np.zeros((self.n_states, self.n_actions))

    def _preprocess(self, state, action):
        if self.continuous_state:
            state = self.state_discretizer.discretize(state)
        if self.continuous_action:
            action = self.action_discretizer.discretize(action)
        return state, action

    def reset(self):
        self.N_sa = np.zeros((self.n_states, self.n_actions))

    def update(self, state, action, next_state, reward, **kwargs):
        state, action = self._preprocess(state, action)
        self.N_sa[state, action] += 1

    def measure(self, state, action, **kwargs):
        state, action = self._preprocess(state, action)
        n = np.maximum(1.0, self.N_sa[state, action])
        return 1.0/np.sqrt(n)

    def count(self, state, action):
        state, action = self._preprocess(state, action)
        return self.N_sa[state, action]



"""
This files provides a set of 2D environments with increasing difficulty
of exploration.

The difficulty is ranked by the level.

Important:
    * To create instances, use the function get_benchmark_env(level).
    * The horizon H is also set as an attribute of the environment.
"""

import numpy as np

from rlberry.wrappers.autoreset import AutoResetWrapper
from rlberry.envs.benchmarks.ball_exploration.pball import PBall2D


def get_benchmark_env(level=1):
    if level == 1:
        env = _get_autoreset_env(BallLevel1())
        return env
    elif level == 2:
        env = _get_autoreset_env(BallLevel2())
        return env
    elif level == 3:
        env = _get_autoreset_env(BallLevel3())
        return env
    elif level == 4:
        env = _get_autoreset_env(BallLevel4())
        return env
    elif level == 5:
        env = _get_autoreset_env(BallLevel5())
        return env
    else:
        raise NotImplementedError("Invalid benchmark level.")


def _get_autoreset_env(env):
    horizon = env.horizon
    return AutoResetWrapper(env, horizon)


#
# Level 1
#

class BallLevel1(PBall2D):
    """
    Dense rewards
    """
    def __init__(self):
        self.horizon = 30
        #
        self.p = 2
        self.action_list = [np.array([0.0, 0.0]),
                            0.05 * np.array([1.0, 0.0]),
                            -0.05 * np.array([1.0, 0.0]),
                            0.05 * np.array([0.0, 1.0]),
                            -0.05 * np.array([0.0, 1.0])]

        self.reward_amplitudes = np.array([1.0])
        self.reward_smoothness = np.array([0.5*np.sqrt(2)])
        self.reward_centers = [np.array([0.5, 0.5])]
        self.A = np.eye(2)
        self.B = np.eye(2)
        self.sigma = 0.01
        self.sigma_init = 0.001
        self.mu_init = np.array([0.0, 0.0])

        PBall2D.__init__(self,
                         self.p,
                         self.action_list,
                         self.reward_amplitudes,
                         self.reward_smoothness,
                         self.reward_centers,
                         self.A,
                         self.B,
                         self.sigma,
                         self.sigma_init,
                         self.mu_init)
        self.name = "Ball Exploration Benchmark - Level 1"


#
# Level 2
#

class BallLevel2(BallLevel1):
    """
    Sparse rewards
    """
    def __init__(self):
        BallLevel1.__init__(self)
        self.reward_amplitudes = np.array([1.0])
        self.reward_smoothness = np.array([0.2])
        self.reward_centers = [np.array([0.5, 0.5])]
        self.name = "Ball Exploration Benchmark - Level 2"

#
# Level 3
#


class BallLevel3(BallLevel2):
    """
    Sparse rewards, noisier
    """
    def __init__(self):
        BallLevel2.__init__(self)
        self.sigma = 0.025
        self.name = "Ball Exploration Benchmark - Level 3"


#
# Level 4
#


class BallLevel4(BallLevel1):
    """
    Far sparse reward (as lvl 2) + dense suboptimal rewards
    """
    def __init__(self):
        BallLevel1.__init__(self)

        self.reward_amplitudes = np.array([1.0, 0.1])
        self.reward_smoothness = np.array([0.2, 0.5*np.sqrt(2)])
        self.reward_centers = [np.array([-0.5, -0.5]),     # far sparse
                               np.array([0.5, 0.5])]       # dense
        self.name = "Ball Exploration Benchmark - Level 4"


#
# Level 5
#

class BallLevel5(BallLevel4):
    """
    Far sparse reward (as lvl 2) + dense suboptimal rewards, noisier
    """
    def __init__(self):
        BallLevel4.__init__(self)
        self.sigma = 0.025
        self.name = "Ball Exploration Benchmark - Level 5"


# if __name__ == '__main__':
#     env = get_benchmark_env(1)
#     env.enable_rendering()
#     for ii in range(100):
#         # env.step(1)
#         # env.step(3)
#         # env.step(env.action_space.sample())
#         # env.step(0)
#         env.step(4)

#     env.render()



from rlberry.wrappers import Wrapper


class AutoResetWrapper(Wrapper):
    """
    Auto reset the environment after "horizon" steps have passed.
    """
    def __init__(self, env, horizon):
        """
        Parameters
        ----------
        horizon: int
        """
        Wrapper.__init__(self, env)
        self.horizon = horizon
        assert self.horizon >= 1
        self.current_step = 0

    def reset(self):
        self.current_step = 0
        return self.env.reset()

    def step(self, action):
        observation, reward, done, info = self.env.step(action)
        self.current_step += 1
        # At H, always return to the initial state.
        # Also, set done to True.
        if self.current_step == self.horizon:
            self.current_step = 0
            observation = self.env.reset()
            done = True
        return observation, reward, done, info



import numpy as np

from rlberry.spaces import Space


class Discrete(Space):
    """
    Class that represents discrete spaces.

    Attributes
    ----------
    n : int
        number of elements in the space
    rng : numpy.random._generator.Generator
        random number generator provided by rlberry.seeding

    Methods
    -------
    sample()
        sample element from the space
    contains(x)
        check if x belongs to the space
    """

    def __init__(self, n):
        """
        Parameters
        ----------
        n : int
            number of elements in the space
        """
        assert n >= 0, "The number of elements in Discrete must be >= 0"
        super(Discrete, self).__init__()
        self.n = n

    def sample(self):
        return self.rng.integers(0, self.n)

    def contains(self, x):
        return (x >= 0) and (x < self.n) and np.issubdtype(type(x), np.integer)

    def __str__(self):
        objstr = "%d-element Discrete space" % self.n
        return objstr



import numpy as np
from rlberry.utils.jit_setup import numba_jit


@numba_jit
def kernel_func(z, kernel_type):
    """
    Returns a kernel function to the real value z.

    Kernel types:

    "uniform"      : 1.0*(abs(z) <= 1)
    "triangular"   : max(0, 1 - abs(z))
    "gaussian"     : exp(-z^2/2)
    "epanechnikov" : max(0, 1-z^2)
    "quartic"      : (1-z^2)^2 *(abs(z) <= 1)
    "triweight"    : (1-z^2)^3 *(abs(z) <= 1)
    "tricube"      : (1-abs(z)^3)^3 *(abs(z) <= 1)
    "cosine"       : cos( z * (pi/2) ) *(abs(z) <= 1)

    Parameters
    ----------
    z : double
    kernel_type : string
    """
    if kernel_type == "uniform":
        return 1.0 * (np.abs(z) <= 1)
    elif kernel_type == "triangular":
        return (1.0 - np.abs(z)) * (np.abs(z) <= 1)
    elif kernel_type == "gaussian":
        return np.exp(-np.power(z, 2.0) / 2.0)
    elif kernel_type == "epanechnikov":
        return (1.0 - np.power(z, 2.0)) * (np.abs(z) <= 1)
    elif kernel_type == "quartic":
        return np.power((1.0 - np.power(z, 2.0)), 2.0)*(np.abs(z) <= 1)
    elif kernel_type == "triweight":
        return np.power((1.0 - np.power(z, 2.0)), 3.0)*(np.abs(z) <= 1)
    elif kernel_type == "tricube":
        return np.power((1.0 - np.power(np.abs(z), 3.0)), 3.0)*(np.abs(z) <= 1)
    elif kernel_type == "cosine":
        return np.cos(z*np.pi/2)*(np.abs(z) <= 1)
    else:
        raise NotImplementedError("Unknown kernel type.")



import time

import numpy as np
import torch
import torch.nn as nn
from torch.distributions import Categorical

import rlberry.spaces as spaces
from rlberry.agents import IncrementalAgent

# choose device
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class ValueNet(nn.Module):
    def __init__(self, state_dim, hidden_size=64):
        super(ValueNet, self).__init__()
        # critic
        self.critic = nn.Sequential(
            nn.Linear(state_dim, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, 1)
        )

    def forward(self, state):
        state_value = self.critic(state)
        return torch.squeeze(state_value)


class PolicyNet(nn.Module):
    def __init__(self, state_dim, action_dim, hidden_size=64):
        super(PolicyNet, self).__init__()
        # actor
        self.actor = nn.Sequential(
            nn.Linear(state_dim, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, action_dim)
        )
        self.softmax = nn.Softmax(dim=-1)

    def forward(self, state):
        action_probs = self.softmax(self.actor(state))
        dist = Categorical(action_probs)
        return dist

    def action_scores(self, state):
        action_scores = self.actor(state)
        return action_scores


class Memory:
    def __init__(self):
        self.actions = []
        self.states = []
        self.logprobs = []
        self.rewards = []
        self.is_terminals = []

    def clear_memory(self):
        del self.actions[:]
        del self.states[:]
        del self.logprobs[:]
        del self.rewards[:]
        del self.is_terminals[:]


class PPOAgent(IncrementalAgent):
    """
    References
    ----------
    Schulman, J., Wolski, F., Dhariwal, P., Radford, A. & Klimov, O. (2017).
    "Proximal Policy Optimization Algorithms."
    arXiv preprint arXiv:1707.06347.

    Schulman, J., Levine, S., Abbeel, P., Jordan, M., & Moritz, P. (2015).
    "Trust region policy optimization."
    In International Conference on Machine Learning (pp. 1889-1897).
    """

    name = "PPO"
    fit_info = ("n_episodes", "episode_rewards")

    def __init__(self, env,
                 n_episodes=4000,
                 batch_size=8,
                 horizon=256,
                 gamma=0.99,
                 learning_rate=0.01,
                 eps_clip=0.2,
                 k_epochs=5,
                 verbose=1,
                 **kwargs):
        """
        env : Model
            Online model with continuous (Box) state space and discrete actions
        n_episodes : int
            Number of episodes
        batch_size : int
            Number of episodes to wait before updating the policy.
        horizon : int
            Horizon.
        gamma : double
            Discount factor in [0, 1].
        learning_rate : double
            Learning rate.
        eps_clip : double
            PPO clipping range (epsilon).
        k_epochs : int
            Number of epochs per update.
        verbose : int
            Controls the verbosity, if non zero, progress messages are printed.
        """
        IncrementalAgent.__init__(self, env, **kwargs)

        self.n_episodes = n_episodes
        self.batch_size = batch_size
        self.horizon = horizon
        self.gamma = gamma
        self.learning_rate = learning_rate
        self.eps_clip = eps_clip
        self.k_epochs = k_epochs

        self.state_dim = self.env.observation_space.dim
        self.action_dim = self.env.action_space.n
        self.verbose = verbose

        # check environment
        assert isinstance(self.env.observation_space, spaces.Box)
        assert isinstance(self.env.action_space, spaces.Discrete)

        self.cat_policy = None  # categorical policy function

        # initialize
        self.reset()

    def reset(self, **kwargs):
        self.cat_policy = PolicyNet(self.state_dim, self.action_dim).to(device)
        self.policy_optimizer = torch.optim.Adam(self.cat_policy.parameters(),
                                                 lr=self.learning_rate,
                                                 betas=(0.9, 0.999))

        self.value_net = ValueNet(self.state_dim).to(device)
        self.value_optimizer = torch.optim.Adam(self.value_net.parameters(),
                                                lr=self.learning_rate,
                                                betas=(0.9, 0.999))

        self.cat_policy_old = \
            PolicyNet(self.state_dim, self.action_dim).to(device)
        self.cat_policy_old.load_state_dict(self.cat_policy.state_dict())

        self.MseLoss = nn.MSELoss()

        self.memory = Memory()

        self.episode = 0

        # useful data
        self._rewards = np.zeros(self.n_episodes)
        self._cumul_rewards = np.zeros(self.n_episodes)

        # logging config
        self._last_printed_ep = 0
        self._time_last_log = time.process_time()
        if self.verbose == 1:
            self._log_interval = 60  # in seconds
        elif self.verbose == 2:
            self._log_interval = 30
        elif self.verbose == 3:
            self._log_interval = 15
        elif self.verbose > 3:
            self._log_interval = 5

    def policy(self, state, **kwargs):
        assert self.cat_policy is not None
        state = torch.from_numpy(state).float().to(device)
        action_dist = self.cat_policy_old(state)
        action = action_dist.sample().item()
        return action

    def fit(self, **kwargs):
        info = {}
        for k in range(self.n_episodes):
            self._run_episode()

        info["n_episodes"] = self.n_episodes
        info["episode_rewards"] = self._rewards
        return info

    def partial_fit(self, fraction, **kwargs):
        assert fraction > 0.0 and fraction <= 1.0
        n_episodes_to_run = int(np.ceil(fraction*self.n_episodes))
        count = 0
        while count < n_episodes_to_run and self.episode < self.n_episodes:
            self._run_episode()
            count += 1

        info = {}
        info["n_episodes"] = self.episode
        info["episode_rewards"] = self._rewards[:self.episode]
        return info

    def _logging(self):
        if self.verbose > 0:
            t_now = time.process_time()
            time_elapsed = t_now - self._time_last_log
            if time_elapsed >= self._log_interval:
                self._time_last_log = t_now
                print(self._info_to_print())
                self._last_printed_ep = self.episode - 1

    def _info_to_print(self):
        prev_episode = self._last_printed_ep
        episode = self.episode - 1
        reward_per_ep = self._rewards[prev_episode:episode + 1].sum() \
            / max(1, episode - prev_episode)
        time_per_ep = self._log_interval * 1000.0 \
            / max(1, episode - prev_episode)
        time_per_ep = max(0.01, time_per_ep)  # avoid div by zero
        fps = int((self.horizon / time_per_ep) * 1000)

        to_print = "[{}] episode = {}/{} ".format(self.name, episode+1,
                                                  self.n_episodes) \
            + "| reward/ep = {:0.2f} ".format(reward_per_ep) \
            + "| time/ep = {:0.2f} ms".format(time_per_ep) \
            + "| fps = {}".format(fps)
        return to_print

    def _select_action(self, state):
        state = torch.from_numpy(state).float().to(device)
        action_dist = self.cat_policy_old(state)
        action = action_dist.sample()
        action_logprob = action_dist.log_prob(action)

        self.memory.states.append(state)
        self.memory.actions.append(action)
        self.memory.logprobs.append(action_logprob)

        return action.item()

    def _run_episode(self):
        # interact for H steps
        episode_rewards = 0
        state = self.env.reset()
        for t in range(self.horizon):
            # running policy_old
            action = self._select_action(state)
            next_state, reward, done, _ = self.env.step(action)

            # save in batch
            self.memory.rewards.append(reward)
            self.memory.is_terminals.append(done)
            episode_rewards += reward

            if done:
                break

            # update state
            state = next_state

        # update
        ep = self.episode
        self._rewards[ep] = episode_rewards
        self._cumul_rewards[ep] = episode_rewards \
            + self._cumul_rewards[max(0, ep - 1)]
        self.episode += 1
        self._logging()

        #
        if self.episode % self.batch_size == 0:
            self._update()
            self.memory.clear_memory()

        return episode_rewards

    def _update(self):
        # monte carlo estimate of rewards
        rewards = []
        discounted_reward = 0
        for reward, is_terminal in zip(reversed(self.memory.rewards),
                                       reversed(self.memory.is_terminals)):
            if is_terminal:
                discounted_reward = 0
            discounted_reward = reward + (self.gamma * discounted_reward)
            rewards.insert(0, discounted_reward)

        # normalizing the rewards
        rewards = torch.tensor(rewards).to(device).float()
        rewards = (rewards - rewards.mean()) / (rewards.std() + 1e-5)

        # convert list to tensor
        old_states = torch.stack(self.memory.states).to(device).detach()
        old_actions = torch.stack(self.memory.actions).to(device).detach()
        old_logprobs = torch.stack(self.memory.logprobs).to(device).detach()

        # optimize policy for K epochs
        for _ in range(self.k_epochs):
            # evaluate old actions and values
            action_dist = self.cat_policy(old_states)
            logprobs = action_dist.log_prob(old_actions)
            state_values = self.value_net(old_states)
            dist_entropy = action_dist.entropy()

            # find ratio (pi_theta / pi_theta__old)
            ratios = torch.exp(logprobs - old_logprobs.detach())

            # find surrogate loss
            advantages = rewards - state_values.detach()
            surr1 = ratios * advantages
            surr2 = torch.clamp(ratios, 1 - self.eps_clip,
                                1 + self.eps_clip) * advantages
            loss = -torch.min(surr1, surr2) \
                + 0.5 * self.MseLoss(state_values, rewards) \
                - 0.01 * dist_entropy

            # take gradient step
            self.policy_optimizer.zero_grad()
            self.value_optimizer.zero_grad()

            loss.mean().backward()

            self.policy_optimizer.step()
            self.value_optimizer.step()

        # copy new weights into old policy
        self.cat_policy_old.load_state_dict(self.cat_policy.state_dict())

    #
    # For hyperparameter optimization
    #
    @classmethod
    def sample_parameters(cls, trial):
        batch_size = trial.suggest_categorical('batch_size',
                                               [1, 4, 8, 16, 32, 64])
        learning_rate = trial.suggest_loguniform('learning_rate', 1e-5, 1)
        return {
                'batch_size': batch_size,
                'learning_rate': learning_rate,
                }



from numpy.random import SeedSequence, default_rng

# Define global seed sequence
_GLOBAL_SEED = None
_GLOBAL_SEED_SEQ = None


#
# Seeding other libraries
#

_TORCH_INSTALLED = True

try:
    import torch
except Exception:
    _TORCH_INSTALLED = False


#
# Seeding functions
#

def set_global_seed(seed=42):
    """
    rlberry has a global seed from which we can obtain different random number
    generators with (close to) independent outcomes.

    Important:

    If the global seed is altered by the user, it should be done only once,
    after importing rlberry for the first time. This is to ensure that all
    random number generators are children of the same SeedSequence.

    To do:
        Check (torch seeding):
        https://github.com/pytorch/pytorch/issues/7068#issuecomment-487907668
    """
    global _GLOBAL_SEED, _GLOBAL_SEED_SEQ
    _GLOBAL_SEED = seed
    _GLOBAL_SEED_SEQ = SeedSequence(_GLOBAL_SEED)

    # get state (for seeding)
    # rng_libs = get_rng()
    # state = rng_libs.__getstate__()['state']['state']

    # seed torch
    if _TORCH_INSTALLED:
        torch.manual_seed(seed)


def get_rng():
    """
    Get a random number generator (rng), from the global seed sequence.

    Returns
    -------
    rng : numpy.random._generator.Generator
        random number generator
    """
    # Spawn off 1 child SeedSequence
    child_seed = _GLOBAL_SEED_SEQ.spawn(1)
    rng = default_rng(child_seed[0])
    return rng


#
# Set global seed
#

set_global_seed()



import numpy as np
import os
import rlberry.seeding as seeding
from rlberry.envs import GridWorld
from rlberry.agents import IncrementalAgent
from rlberry.stats import AgentStats, plot_episode_rewards, compare_policies


# global seed
seeding.set_global_seed(1234)


class DummyAgent(IncrementalAgent):
    fit_info = ("episode_rewards",)

    def __init__(self, env, n_episodes, hyperparameter=0, **kwargs):
        IncrementalAgent.__init__(self, env, **kwargs)
        self.name = "DummyAgent"
        self.n_episodes = n_episodes
        self.fitted = False
        self.hyperparameter = hyperparameter

        self.fraction_fitted = 0.0

    def fit(self, **kwargs):
        info = {}
        info["episode_rewards"] = np.arange(self.n_episodes)
        self.fitted = True
        return info

    def partial_fit(self, fraction, **kwargs):
        assert fraction > 0.0 and fraction <= 1.0
        self.fraction_fitted = min(1.0, self.fraction_fitted + fraction)
        info = {}
        nn = int(np.ceil(fraction*self.n_episodes))
        info["episode_rewards"] = np.arange(nn)
        return info

    def policy(self, observation, time=0, **kwargs):
        return self.env.action_space.sample()

    @classmethod
    def sample_parameters(cls, trial):
        hyperparameter = trial.suggest_categorical('hyperparameter', [1, 2, 3])
        return {'hyperparameter': hyperparameter}


def test_agent_stats_1():
    # Define train and evaluation envs
    train_env = GridWorld()
    eval_env = GridWorld()

    # Parameters
    params = {"n_episodes": 500, "horizon": 20}

    # Check DummyAgent
    agent = DummyAgent(train_env, **params)
    agent.fit()
    agent.policy(None)

    # Run AgentStats
    stats_agent1 = AgentStats(DummyAgent, train_env,
                              init_kwargs=params, n_fit=4, eval_horizon=10)
    stats_agent2 = AgentStats(DummyAgent, train_env,
                              init_kwargs=params, n_fit=4, eval_horizon=10)
    agent_stats_list = [stats_agent1, stats_agent2]

    # learning curves
    plot_episode_rewards(agent_stats_list, cumulative=True, show=False)

    # compare final policies
    compare_policies(agent_stats_list, eval_env,
                     eval_horizon=params["horizon"], n_sim=10, show=False)
    compare_policies(agent_stats_list, eval_env,
                     eval_horizon=params["horizon"],
                     n_sim=10, show=False, stationary_policy=False)

    # check if fitted
    for agent_stats in agent_stats_list:
        assert len(agent_stats.fitted_agents) == 4
        for agent in agent_stats.fitted_agents:
            assert agent.fitted

    # test saving/loading
    stats_agent1.save('test_agent_stats_file.pickle')
    loaded_stats = AgentStats.load('test_agent_stats_file.pickle')
    assert stats_agent1.identifier == loaded_stats.identifier

    # delete file
    os.remove('test_agent_stats_file.pickle')

    # test hyperparemeter optimization
    loaded_stats.optimize_hyperparams()
    loaded_stats.optimize_hyperparams(continue_previous=True)


def test_agent_stats_2():
    # Define train and evaluation envs
    train_env = GridWorld()
    eval_env = GridWorld()

    # Parameters
    params = {"n_episodes": 500, "horizon": 20}

    # Run AgentStats
    stats_agent1 = AgentStats(DummyAgent, train_env, eval_env=eval_env,
                              init_kwargs=params, n_fit=4, eval_horizon=10,
                              n_jobs=1)
    stats_agent2 = AgentStats(DummyAgent, train_env, eval_env=eval_env,
                              init_kwargs=params, n_fit=4, eval_horizon=10,
                              n_jobs=1)
    agent_stats_list = [stats_agent1, stats_agent2]

    # compare final policies
    compare_policies(agent_stats_list, n_sim=10, show=False)
    compare_policies(agent_stats_list,
                     n_sim=10, show=False, stationary_policy=False)

    # learning curves
    plot_episode_rewards(agent_stats_list, cumulative=True, show=False)

    # check if fitted
    for agent_stats in agent_stats_list:
        assert len(agent_stats.fitted_agents) == 4
        for agent in agent_stats.fitted_agents:
            assert agent.fitted

    # test saving/loading
    stats_agent1.save()
    loaded_stats = AgentStats.load(stats_agent1.default_filename)
    assert stats_agent1.identifier == loaded_stats.identifier

    # delete file
    os.remove(stats_agent1.default_filename+'.pickle')

    # test hyperparemeter optimization
    loaded_stats.optimize_hyperparams()


def test_agent_stats_partial_fit():
    # Define train and evaluation envs
    train_env = GridWorld()
    eval_env = GridWorld()

    # Parameters
    params = {"n_episodes": 500, "horizon": 20}

    # Check DummyAgent
    agent = DummyAgent(train_env, **params)
    agent.fit()
    agent.policy(None)

    # Run AgentStats
    stats = AgentStats(DummyAgent, train_env,
                       init_kwargs=params, n_fit=4, eval_horizon=10)

    # Run partial fit
    stats.partial_fit(0.1)
    stats.partial_fit(0.5)
    for agent in stats.fitted_agents:
        assert agent.fraction_fitted == 0.6
    for _ in range(2):
        stats.partial_fit(0.5)
        for agent in stats.fitted_agents:
            assert agent.fraction_fitted == 1.0

    # learning curves
    plot_episode_rewards([stats], cumulative=True, show=False)

    # compare final policies
    compare_policies([stats], eval_env,
                     eval_horizon=params["horizon"], n_sim=10, show=False)



from abc import ABC, abstractmethod


class UncertaintyEstimator(ABC):
    def __init__(self, observation_space, action_space, **kwargs):
        super().__init__()
        self.observation_space = observation_space
        self.action_space = action_space

    def reset(self, **kwargs):
        pass

    @abstractmethod
    def update(self, state, action, next_state, reward, **kwargs):
        pass

    @abstractmethod
    def measure(self, state, action, **kwargs):
        pass



import time

import numpy as np
import torch
import torch.nn as nn
from torch.distributions import Categorical

import rlberry.spaces as spaces
from rlberry.agents import Agent

# choose device
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class ActorCritic(nn.Module):
    def __init__(self, state_dim, action_dim):
        super(ActorCritic, self).__init__()
        # actor
        self.actor = nn.Sequential(
            nn.Linear(state_dim, 64),
            nn.Tanh(),
            nn.Linear(64, 64),
            nn.Tanh(),
            nn.Linear(64, action_dim),
            nn.Softmax(dim=-1)
        )
        # critic
        self.critic = nn.Sequential(
            nn.Linear(state_dim, 64),
            nn.Tanh(),
            nn.Linear(64, 64),
            nn.Tanh(),
            nn.Linear(64, 1)
        )

    def forward(self):
        raise NotImplementedError

    def act(self, state):
        action_probs = self.actor(state)
        dist = Categorical(action_probs)
        action = dist.sample()

        return action, dist.log_prob(action)

    def evaluate(self, state, action):
        action_probs = self.actor(state)
        dist = Categorical(action_probs)

        action_logprobs = dist.log_prob(action)
        dist_entropy = dist.entropy()

        state_value = self.critic(state)

        return action_logprobs, torch.squeeze(state_value), dist_entropy


class Memory:
    def __init__(self):
        self.actions = []
        self.states = []
        self.logprobs = []
        self.rewards = []
        self.is_terminals = []

    def clear_memory(self):
        del self.actions[:]
        del self.states[:]
        del self.logprobs[:]
        del self.rewards[:]
        del self.is_terminals[:]


class AVECPPOAgent(Agent):
    """
    AVEC uses a modification of the training objective for the critic in
    actor-critic algorithms to better approximate the value function (critic).
    The new state-value function approximation learns the *relative* value of
    the states rather than their *absolute* value as in conventional
    actor-critic. This modification is:
    - well-motivated by recent studies [1,2];
    - theoretically sound;
    - intuitively supported by the need to improve the approximation error
    of the critic.

    The application of Actor with Variance Estimated Critic (AVEC) to
    state-of-the-art policy gradient methods produces considerable
    gains in performance (on average +26% for SAC and +40% for PPO)
    over the standard actor-critic training.

    References
    ----------
    Flet-Berliac, Y., Ouhamma, R., Maillard, O. A., & Preux, P. (2020).
    "Is Standard Deviation the New Standard? Revisiting the Critic in Deep
    Policy Gradients."
    arXiv preprint arXiv:2010.04440.

    [1] Ilyas, A., Engstrom, L., Santurkar, S., Tsipras, D., Janoos, F.,
    Rudolph, L. & Madry, A. (2020).
    "A closer look at deep policy gradients."
    In International Conference on Learning Representations.

    [2] Tucker, G., Bhupatiraju, S., Gu, S., Turner, R., Ghahramani, Z. &
    Levine, S. (2018).
    "The mirage of action-dependent baselines in reinforcement learning."
    In International Conference on Machine Learning, pp. 5015–5024.
    """

    name = "AVECPPO"
    fit_info = ("n_episodes", "episode_rewards")

    def __init__(self, env,
                 n_episodes=4000,
                 horizon=256,
                 gamma=0.99,
                 lr=0.0003,
                 eps_clip=0.2,
                 k_epochs=10,
                 verbose=1,
                 **kwargs):
        """
        env : Model
            model with continuous (Box) state space and discrete actions
        n_episodes : int
            Number of episodes
        horizon : int
            Horizon of the objective function. If None and gamma<1,
            set to 1/(1-gamma).
        gamma : double
            Discount factor in [0, 1]. If gamma is 1.0, the problem is set
            to be finite-horizon.
        lr : double
            Learning rate.
        eps_clip : double
            PPO clipping range (epsilon).
        k_epochs : int
            Number of epochs per update.
        verbose : int
            Controls the verbosity, if non zero, progress messages are printed.
        """
        Agent.__init__(self, env, **kwargs)

        self.lr = lr
        self.gamma = gamma
        self.eps_clip = eps_clip
        self.k_epochs = k_epochs
        self.horizon = horizon
        self.n_episodes = n_episodes
        self.state_dim = self.env.observation_space.dim
        self.action_dim = self.env.action_space.n
        self.verbose = verbose

        # check environment
        assert self.env.is_online()
        assert isinstance(self.env.observation_space, spaces.Box)
        assert isinstance(self.env.action_space, spaces.Discrete)

        self.cat_policy = None  # categorical policy function

        # initialize
        self.reset()

    def reset(self, **kwargs):
        self.cat_policy = ActorCritic(self.state_dim,
                                      self.action_dim).to(device)
        self.optimizer = torch.optim.Adam(self.cat_policy.parameters(),
                                          lr=self.lr, betas=(0.9, 0.999))

        self.cat_policy_old = ActorCritic(self.state_dim,
                                          self.action_dim).to(device)
        self.cat_policy_old.load_state_dict(self.cat_policy.state_dict())

        self.memory = Memory()

        self.episode = 0

        # logging config
        self._last_printed_ep = 0
        self._time_last_log = time.process_time()
        if self.verbose == 1:
            self._log_interval = 60  # in seconds
        elif self.verbose == 2:
            self._log_interval = 30
        elif self.verbose == 3:
            self._log_interval = 15
        elif self.verbose > 3:
            self._log_interval = 5

    def policy(self, state, **kwargs):
        assert self.cat_policy is not None

        return self._select_action(state)

    def fit(self, **kwargs):
        info = {}
        self._rewards = np.zeros(self.n_episodes)
        self._cumul_rewards = np.zeros(self.n_episodes)
        for k in range(self.n_episodes):
            episode_rewards = self._run_episode()
            self._rewards[k] = episode_rewards
            if k > 0:
                self._cumul_rewards[k] = episode_rewards \
                    + self._cumul_rewards[k - 1]
            self.episode += 1
            self._logging()

            # update
            self._update()
            self.memory.clear_memory()

        info["n_episodes"] = self.n_episodes
        info["episode_rewards"] = self._rewards
        return info

    def _logging(self):
        if self.verbose > 0:
            t_now = time.process_time()
            time_elapsed = t_now - self._time_last_log
            if time_elapsed >= self._log_interval:
                self._time_last_log = t_now
                print(self._info_to_print())
                self._last_printed_ep = self.episode - 1

    def _info_to_print(self):
        prev_episode = self._last_printed_ep
        episode = self.episode - 1
        reward_per_ep = self._rewards[prev_episode:episode + 1].sum() / \
            max(1, episode - prev_episode)
        time_per_ep = self._log_interval * 1000.0 / \
            max(1, episode - prev_episode)
        time_per_ep = max(0.01, time_per_ep)  # avoid div by zero
        fps = int((self.horizon / time_per_ep) * 1000)

        to_print = "[{}] episode = {}/{} ".format(self.name, episode+1,
                                                  self.n_episodes) \
            + "| reward/ep = {:0.2f} ".format(reward_per_ep) \
            + "| time/ep = {:0.2f} ms".format(time_per_ep) \
            + "| fps = {}".format(fps)
        return to_print

    def _select_action(self, state):
        state = torch.from_numpy(state).float().to(device)
        action, action_logprob = self.cat_policy_old.act(state)

        self.memory.states.append(state)
        self.memory.actions.append(action)
        self.memory.logprobs.append(action_logprob)

        return action.item()

    def _run_episode(self):
        # interact for H steps
        episode_rewards = 0
        state = self.env.reset()
        for _ in range(self.horizon):
            # running policy_old
            action = self._select_action(state)
            state, reward, done, _ = self.env.step(action)

            # save in batch
            self.memory.rewards.append(reward)
            self.memory.is_terminals.append(done)
            episode_rewards += reward

            if done:
                break

        return episode_rewards

    def _update(self):
        # monte carlo estimate of rewards
        rewards = []
        discounted_reward = 0
        for reward, is_terminal in zip(reversed(self.memory.rewards),
                                       reversed(self.memory.is_terminals)):
            if is_terminal:
                discounted_reward = 0
            discounted_reward = reward + (self.gamma * discounted_reward)
            rewards.insert(0, discounted_reward)

        # normalizing the rewards
        rewards = torch.tensor(rewards).to(device).float()
        rewards = (rewards - rewards.mean()) / (rewards.std() + 1e-5)

        # convert list to tensor
        old_states = torch.stack(self.memory.states).to(device).detach()
        old_actions = torch.stack(self.memory.actions).to(device).detach()
        old_logprobs = torch.stack(self.memory.logprobs).to(device).detach()

        # optimize policy for K epochs
        for _ in range(self.k_epochs):
            # evaluate old actions and values
            logprobs, state_values, dist_entropy = \
                self.cat_policy.evaluate(old_states, old_actions)

            # find ratio (pi_theta / pi_theta__old)
            ratios = torch.exp(logprobs - old_logprobs.detach())

            # find surrogate loss
            advantages = rewards - state_values.detach()
            surr1 = ratios * advantages
            surr2 = torch.clamp(ratios, 1 - self.eps_clip, 1
                                + self.eps_clip) * advantages
            loss = -torch.min(surr1, surr2) \
                + 0.5 * self._avec_loss(state_values, rewards) \
                - 0.01 * dist_entropy

            # take gradient step
            self.optimizer.zero_grad()
            loss.mean().backward()
            self.optimizer.step()

        # copy new weights into old policy
        self.cat_policy_old.load_state_dict(self.cat_policy.state_dict())

    def _avec_loss(self, y_pred, y_true):
        """
        Computes the objective function used in AVEC for the learning
        of the value function:
        the residual variance between the state-values and the
        empirical returns.

        Returns Var[y-ypred]
        :param y_pred: (np.ndarray) the prediction
        :param y_true: (np.ndarray) the expected value
        :return: (float) residual variance of ypred and y
        """
        assert y_true.ndim == 1 and y_pred.ndim == 1

        return torch.var(y_true - y_pred)



import os
import pytest

from pyvirtualdisplay import Display
from rlberry.envs.classic_control import MountainCar
from rlberry.envs.classic_control import Acrobot
from rlberry.envs.finite import Chain
from rlberry.envs.finite import GridWorld
from rlberry.envs.benchmarks.ball_exploration import PBall2D, SimplePBallND
from rlberry.rendering import RenderInterface
from rlberry.rendering import RenderInterface2D

try:
    display = Display(visible=0, size=(1400, 900))
    display.start()
except Exception:
    pass

classes = [
    Acrobot,
    MountainCar,
    GridWorld,
    Chain,
    PBall2D,
    SimplePBallND
]


@pytest.mark.parametrize("ModelClass", classes)
def test_instantiation(ModelClass):
    env = ModelClass()

    if isinstance(env, RenderInterface):
        env.disable_rendering()
        assert not env.is_render_enabled()
        env.enable_rendering()
        assert env.is_render_enabled()


@pytest.mark.parametrize("ModelClass", classes)
def test_render2d_interface(ModelClass):
    env = ModelClass()
    env._debug_mode = True       # rendering debug

    if isinstance(env, RenderInterface2D):
        env.enable_rendering()

        if env.is_online():
            for _ in range(2):
                state = env.reset()
                for _ in range(5):
                    assert env.observation_space.contains(state)
                    action = env.action_space.sample()
                    next_s, _, _, _ = env.step(action)
                    state = next_s
                env.render()
            env.save_video('test_video.mp4')
            env.clear_render_buffer()
        try:
            os.remove('test_video.mp4')
        except Exception:
            pass




from abc import ABC, abstractmethod

from rlberry.seeding import seeding


class Space(ABC):
    """
    Base class to define observation and action spaces.

    Attributes
    ----------
    rng : numpy.random._generator.Generator
        random number generator provided by rlberry.seeding

    Methods
    -------
    sample()
        sample element from the space
    contains(x)
        check if x belongs to the space
    """

    def __init__(self):
        super().__init__()
        # random number generator
        self.rng = seeding.get_rng()

    @abstractmethod
    def sample(self):
        """
        Sample one element from the space.
        """
        pass

    @abstractmethod
    def contains(self, x):
        """
        Returns True if x belongs to the space, and False otherwise.
        """
        pass



import numpy as np
import pytest

from rlberry.spaces import Box
from rlberry.spaces import Discrete


@pytest.mark.parametrize("n", list(range(10)))
def test_discrete_space(n):
    sp = Discrete(n)
    for ii in range(n):
        assert sp.contains(ii)

    for ii in range(2 * n):
        assert sp.contains(sp.sample())


@pytest.mark.parametrize("low, high, dim",
                         [
                             (1.0, 10.0, 1),
                             (1.0, 10.0, 2),
                             (1.0, 10.0, 4),
                             (-10.0, 1.0, 1),
                             (-10.0, 1.0, 2),
                             (-10.0, 1.0, 4),
                             (-np.inf, 1.0, 1),
                             (-np.inf, 1.0, 2),
                             (-np.inf, 1.0, 4),
                             (1.0, np.inf, 1),
                             (1.0, np.inf, 2),
                             (1.0, np.inf, 4),
                             (-np.inf, np.inf, 1),
                             (-np.inf, np.inf, 2),
                             (-np.inf, np.inf, 4),
                         ])
def test_box_space_case_1(low, high, dim):
    sp = Box(low, high, dim)
    for ii in range(2 ** dim):
        assert (sp.contains(sp.sample()))


@pytest.mark.parametrize(
    "low, high",
    [
        (np.array([0.0, 0.0, 0.0]), np.array([1.0, 1.0, 1.0])),
        (np.array([-10.0, -10.0, -10.0]), np.array([10.0, 10.0, 10.0])),
        (np.array([-10.0, -10.0, -10.0]), np.array([10.0, 10.0, np.inf])),
        (np.array([-np.inf, -10.0, -10.0]), np.array([10.0, 10.0, np.inf])),
        (np.array([-np.inf, -10.0, -10.0]), np.array([np.inf, 10.0, np.inf]))
    ])
def test_box_space_case_2(low, high):
    sp = Box(low, high)
    if (-np.inf in low) or (np.inf in high):
        assert not sp.is_bounded
    else:
        assert sp.is_bounded
    for ii in range(2 ** sp.dim):
        assert (sp.contains(sp.sample()))



import numpy as np

from rlberry.agents import Agent
from rlberry.agents.dynprog.utils import backward_induction, value_iteration
from rlberry.spaces import Discrete


class MBQVIAgent(Agent):
    """
    Model-Basel Q-Value iteration (MBQVI).

    Builds an empirical MDP and runs value iteration on it.
    Corresponds to the "indirect" algorithm studied by Kearns and Singh (1999).

    Kearns, Michael J., and Satinder P. Singh.
    "Finite-sample convergence rates for Q-learning and indirect algorithms."
    Advances in neural information processing systems. 1999.
    """

    name = "MBQVI"
    fit_info = ("n_samples", "total_samples", "n_iterations", "precision")

    def __init__(self, env,
                 n_samples=10,
                 gamma=0.95,
                 horizon=None,
                 epsilon=1e-6,
                 verbose=1,
                 **kwargs):
        """
        Parameters:
        -----------
        env : Model
            generative model with finite state-action space
        n_samples : int
            number of samples *per state-action pair* used to estimate
            the empirical MDP.
        gamma : double
            discount factor in [0, 1]
        horizon : int
            horizon, if the problem is finite-horizon. if None, the discounted
            problem is solved. default = None
        epsilon : double
            precision of value iteration, only used in discounted problems
            (when horizon is None).
        verbose : int
            controls the verbosity, if non zero, progress messages are printed.
        """
        # initialize base class
        assert env.is_generative(), \
            "MBQVI requires a generative model."
        assert isinstance(env.observation_space, Discrete), \
            "MBQVI requires a finite state space."
        assert isinstance(env.action_space, Discrete), \
            "MBQVI requires a finite action space."
        Agent.__init__(self, env, **kwargs)

        #
        self.n_samples = n_samples
        self.gamma = gamma
        self.horizon = horizon
        self.epsilon = epsilon
        self.verbose = verbose

        # empirical MDP, created in fit()
        self.R_hat = None
        self.P_hat = None

        # value functions
        self.V = None
        self.Q = None

    def _update(self, state, action, next_state, reward):
        """
        Update model statistics.
        """
        self.N_sa[state, action] += 1
        self.N_sas[state, action, next_state] += 1
        self.S_sa[state, action] += reward

    def fit(self, **kwargs):
        """
        Build empirical MDP and run value iteration.
        """
        S = self.env.observation_space.n
        A = self.env.action_space.n
        self.N_sa = np.zeros((S, A))
        self.N_sas = np.zeros((S, A, S))
        self.S_sa = np.zeros((S, A))

        # collect data
        total_samples = S * A * self.n_samples
        count = 0
        if self.verbose > 0:
            print(
                "[{}] collecting {} samples per (s,a)".format(self.name,
                                                              self.n_samples)
                + ", total = {} samples.".format(total_samples)
            )
        for ss in range(S):
            for aa in range(A):
                for nn in range(self.n_samples):
                    next_state, reward, _, _ = self.env.sample(ss, aa)
                    self._update(ss, aa, next_state, reward)

                    count += 1
                    if count % 10000 == 0 and self.verbose > 0:
                        completed = 100 * count / total_samples
                        print("[{}] ... {}/{} ({:0.0f}%)".format(self.name,
                                                                 count,
                                                                 total_samples,
                                                                 completed))

        # build model and run VI
        if self.verbose > 0:
            print("{} building model".format(self.name)
                  + " and running backward induction...")

        N_sa = np.maximum(self.N_sa, 1)
        self.R_hat = self.S_sa / N_sa
        self.P_hat = np.zeros((S, A, S))
        for ss in range(S):
            self.P_hat[:, :, ss] = self.N_sas[:, :, ss] / N_sa

        info = {}
        info["n_samples"] = self.n_samples
        info["total_samples"] = total_samples
        if self.horizon is None:
            assert self.gamma < 1.0, \
                "The discounted setting requires gamma < 1.0"
            self.Q, self.V, n_it = value_iteration(self.R_hat, self.P_hat,
                                                   self.gamma, self.epsilon)
            info["n_iterations"] = n_it
            info["precision"] = self.epsilon
        else:
            self.Q, self.V = backward_induction(self.R_hat, self.P_hat,
                                                self.horizon, self.gamma)
            info["n_iterations"] = self.horizon
            info["precision"] = 0.0
        return info

    def policy(self, state, hh=0, **kwargs):
        """
        Parameters
        -----------
        state : int
        hh : int
            Stage when action is taken (for finite horizon problems,
            the optimal policy depends on hh).
            Not used if horizon is None.
        """
        assert self.env.observation_space.contains(state)
        if self.horizon is None:
            return self.Q[state, :].argmax()
        else:
            assert hh >= 0 and hh < self.horizon
            return self.Q[hh, state, :].argmax()



import numpy as np
import pytest

from rlberry.utils.binsearch import binary_search_nd
from rlberry.utils.binsearch import unravel_index_uniform_bin


def test_binary_search_nd():
    bin1 = np.array([0.0, 1.0, 2.0, 3.0])  # 3 intervals
    bin2 = np.array([1.0, 2.0, 3.0, 4.0])  # 3 intervals
    bin3 = np.array([2.0, 3.0, 4.0, 5.0, 6.0])  # 4 intervals

    bins = [bin1, bin2, bin3]

    vec1 = np.array([0.0, 1.0, 2.0])
    vec2 = np.array([2.9, 3.9, 5.9])
    vec3 = np.array([1.5, 2.5, 2.5])
    vec4 = np.array([1.5, 2.5, 2.5])

    # index = i + Ni * j + Ni * Nj * k
    assert binary_search_nd(vec1, bins) == 0
    assert binary_search_nd(vec2, bins) == 2 + 3 * 2 + 3 * 3 * 3
    assert binary_search_nd(vec3, bins) == 1 + 3 * 1 + 3 * 3 * 0


@pytest.mark.parametrize("i, j, k, N",
                         [
                             (0, 0, 0, 5),
                             (0, 1, 2, 5),
                             (4, 3, 2, 5),
                             (4, 4, 4, 5)
                         ])
def test_unravel_index_uniform_bin(i, j, k, N):
    # index = i + N * j + N * N * k
    dim = 3
    flat_index = i + N * j + N * N * k
    assert (i, j, k) == unravel_index_uniform_bin(flat_index, dim, N)



from copy import deepcopy
from datetime import datetime
from joblib import Parallel, delayed
import logging
import os
import pickle

import rlberry.seeding as seeding
from rlberry.agents import IncrementalAgent
from rlberry.stats.evaluation import compare_policies

_OPTUNA_INSTALLED = True
try:
    import optuna
except Exception:
    _OPTUNA_INSTALLED = False


#
# Main class
#

class AgentStats:
    """Class to train, optimize hyperparameters, evaluate and gather statistics about an agent."""

    def __init__(self,
                 agent_class,
                 train_env,
                 eval_env=None,
                 eval_horizon=None,
                 init_kwargs=None,
                 fit_kwargs=None,
                 policy_kwargs=None,
                 agent_name=None,
                 n_fit=4,
                 n_jobs=4,
                 output_dir='stats_data',
                 verbose=5):
        """
        Parameters
        ----------
        agent_class
            Class of the agent.
        train_env : Model
            Enviroment used to initialize/train the agent.
        eval_env : Model
            Environment used to evaluate the agent. If None, set to a
            reseeded deep copy of train_env.
        init_kwargs : dict
            Arguments required by the agent's constructor.
        fit_kwargs : dict
            Arguments required to train the agent.
        policy_kwargs : dict
            Arguments required to call agent.policy().
        agent_name : str
            Name of the agent. If None, set to agent_class.name
        n_fit : int
            Number of agent instances to fit.
        n_jobs : int
            Number of jobs to train the agents in parallel using joblib.
        output_dir : str
            Directory where to store data by default.
        verbose : int
            Verbosity level.
        """
        # agent_class should only be None when the constructor is called
        # by the class method AgentStats.load(), since the agent class
        # will be loaded.
        if agent_class is not None:

            self.agent_name = agent_name
            if agent_name is None:
                self.agent_name = agent_class.name

            # create oject identifier
            timestamp = datetime.timestamp(datetime.now())
            self.identifier = 'stats_{}_{}'.format(self.agent_name,
                                                   str(int(timestamp)))

            self.fit_info = agent_class.fit_info
            self.agent_class = agent_class
            self.train_env = train_env
            if eval_env is None:
                self.eval_env = deepcopy(train_env)
                self.eval_env.reseed()
            else:
                self.eval_env = deepcopy(eval_env)
                self.eval_env.reseed()

            self.eval_horizon = eval_horizon
            # init and fit kwargs are deep copied in fit()
            self.init_kwargs = init_kwargs
            self.fit_kwargs = fit_kwargs
            self.policy_kwargs = deepcopy(policy_kwargs)
            self.n_fit = n_fit
            self.n_jobs = n_jobs
            self.output_dir = output_dir
            self.verbose = verbose

            if init_kwargs is None:
                self.init_kwargs = {}
            if fit_kwargs is None:
                self.fit_kwargs = {}
            if policy_kwargs is None:
                self.policy_kwargs = {}

            # Create environment copies for training
            self.train_env_set = []
            for _ in range(n_fit):
                _env = deepcopy(train_env)
                _env.reseed()
                self.train_env_set.append(_env)

            #
            self.fitted_agents = None
            self.fit_kwargs_list = None  # keep in memory for partial_fit()
            self.fit_statistics = {}

            #
            self.rng = seeding.get_rng()

            # optuna study
            self.study = None

            # default filename to save data
            self.default_filename = os.path.join(self.output_dir,
                                                 self.identifier)

    def fit(self):
        """
        Fit the agent instances in parallel.
        """
        if self.verbose > 0:
            print("\n Training AgentStats for %s... \n" % self.agent_name)
        args = [(self.agent_class, train_env,
                deepcopy(self.init_kwargs), deepcopy(self.fit_kwargs))
                for train_env in self.train_env_set]

        workers_output = Parallel(n_jobs=self.n_jobs, verbose=self.verbose)(
            delayed(_fit_worker)(arg) for arg in args)

        self.fitted_agents, stats = (
            [i for i, j in workers_output],
            [j for i, j in workers_output])

        if self.verbose > 0:
            print("\n ... trained! \n")

        # gather all stats in a dictionary
        self._process_fit_statistics(stats)

    def partial_fit(self, fraction):
        """
        Partially fit the agent instances (not parallel).
        """
        assert fraction > 0.0 and fraction <= 1.0
        assert issubclass(self.agent_class, IncrementalAgent)

        # Create instances if this is the first call
        if self.fitted_agents is None:
            self.fitted_agents = []
            self.fit_kwargs_list = []
            for train_env in self.train_env_set:
                init_kwargs = deepcopy(self.init_kwargs)
                agent = self.agent_class(train_env, copy_env=False,
                                         reseed_env=False, **init_kwargs)
                self.fitted_agents.append(agent)
                self.fit_kwargs_list.append(deepcopy(self.fit_kwargs))

        # Run partial fit
        stats = []
        for agent, fit_kwargs in zip(self.fitted_agents, self.fit_kwargs_list):
            info = agent.partial_fit(fraction, **fit_kwargs)
            stats.append(info)
        self._process_fit_statistics(stats)

    def _process_fit_statistics(self, stats):
        """Gather stats in a dictionary"""
        for entry in self.fit_info:
            self.fit_statistics[entry] = []
            for stat in stats:
                self.fit_statistics[entry].append(stat[entry])

    def save(self, filename=None, **kwargs):
        """
        Parameters
        ----------
        filename : string
            Filename with .pickle extension.
            If None, default_filename attribute is used.
        """
        if filename is None:
            filename = self.default_filename
            if not os.path.exists(self.output_dir):
                os.makedirs(self.output_dir)

        if filename[-7:] != '.pickle':
            filename += '.pickle'

        with open(filename, 'wb') as ff:
            pickle.dump(self.__dict__, ff)

    @classmethod
    def load(cls, filename):
        if filename[-7:] != '.pickle':
            filename += '.pickle'

        obj = cls(None, None)
        with open(filename, 'rb') as ff:
            tmp_dict = pickle.load(ff)
        obj.__dict__.clear()
        obj.__dict__.update(tmp_dict)
        return obj

    def optimize_hyperparams(self,
                             n_trials=5,
                             timeout=60,
                             n_sim=5,
                             n_fit=2,
                             n_jobs=2,
                             sampler_method='random',
                             pruner_method='halving',
                             continue_previous=False,
                             partial_fit_fraction=0.25):
        """
        Run hyperparameter optimization and updates init_kwargs with the
        best hyperparameters found.

        Currently supported sampler_method:
            'random'
            'optuna_default'

        Currently supported pruner_method:
            'none'
            'halving'

        Parameters
        ----------
        n_trials: int
            Mumber of agent evaluations
        timeout: int
            Stop study after the given number of second(s).
            Set to None for unlimited time.
        n_sim : int
            Number of Monte Carlo simulations to evaluate a policy.
        n_fit: int
            Number of agents to fit for each hyperparam evaluation.
        n_jobs: int
            Number of jobs to fit agents for each hyperparam evaluation,
            and also the number of jobs of optuna.
        sampler_method : str
            Optuna sampling method.
        pruner_method : str
            Optuna pruner method.
        continue_previous : bool
            Set to true to continue previous optuna study. If true,
            sampler_method and pruner_method will be
            the same as in the previous study.
        partial_fit_fraction : double, in ]0, 1]
            Fraction of the agent to fit for partial evaluation
            (allows pruning of trials).
            Only used for agents that implement partial_fit()
            (IncrementalAgent interface).
        """
        global _OPTUNA_INSTALLED
        if not _OPTUNA_INSTALLED:
            logging.error("Optuna not installed.")
            return

        assert self.eval_horizon is not None, \
            "To use optimize_hyperparams(), \
eval_horizon must be given to AgentStats."

        assert partial_fit_fraction > 0.0 and partial_fit_fraction <= 1.0

        #
        # Create optuna study
        #
        if continue_previous:
            assert self.study is not None
            study = self.study

        else:
            # get sampler
            if sampler_method == 'random':
                optuna_seed = self.rng.integers(2**16)
                sampler = optuna.samplers.RandomSampler(seed=optuna_seed)
            elif sampler_method == 'optuna_default':
                sampler = None
            else:
                raise NotImplementedError("Sampler method %s is\
 not implemented." % sampler_method)

            # get pruner
            if pruner_method == 'halving':
                pruner = optuna.pruners.SuccessiveHalvingPruner(
                            min_resource=1,
                            reduction_factor=4,
                            min_early_stopping_rate=0)

            elif pruner_method == 'none':
                pruner = None
            else:
                raise NotImplementedError("Pruner method %s is\
 not implemented." % pruner_method)

            # optuna study
            study = optuna.create_study(sampler=sampler,
                                        pruner=pruner,
                                        direction='maximize')
            self.study = study

        def objective(trial):
            kwargs = deepcopy(self.init_kwargs)

            # will raise exception if sample_parameters() is not
            # implemented by the agent class
            kwargs.update(self.agent_class.sample_parameters(trial))

            #
            # fit and evaluate agents
            #
            # Create AgentStats with hyperparams
            params_stats = AgentStats(
                self.agent_class,
                deepcopy(self.train_env),
                init_kwargs=kwargs,   # kwargs are being optimized
                fit_kwargs=deepcopy(self.fit_kwargs),
                policy_kwargs=deepcopy(self.policy_kwargs),
                agent_name='optim',
                n_fit=n_fit,
                n_jobs=n_jobs,
                verbose=0)

            # Evaluation environment copy
            params_eval_env = deepcopy(self.eval_env)
            params_eval_env.reseed()

            #
            # Case 1: partial fit, that allows pruning
            #
            if partial_fit_fraction < 1.0 \
                    and issubclass(params_stats.agent_class, IncrementalAgent):
                fraction_complete = 0.0
                step = 0
                while fraction_complete < 1.0:
                    #
                    params_stats.partial_fit(partial_fit_fraction)
                    # Get rewards
                    eval_result = compare_policies(
                                [params_stats],
                                eval_env=params_eval_env,
                                eval_horizon=self.eval_horizon,
                                stationary_policy=True,
                                n_sim=n_sim,
                                plot=False)

                    rewards = eval_result['optim'].values.mean()
                    # Report intermediate objective value
                    trial.report(rewards, step)

                    #
                    fraction_complete += partial_fit_fraction
                    step += 1
                    #

                    # Handle pruning based on the intermediate value.
                    if trial.should_prune():
                        raise optuna.TrialPruned()

            #
            # Case 2: full fit
            #
            else:
                # Fit and evaluate params_stats
                params_stats.fit()

                # Get rewards
                eval_result = compare_policies(
                            [params_stats],
                            eval_env=params_eval_env,
                            eval_horizon=self.eval_horizon,
                            stationary_policy=True,
                            n_sim=n_sim,
                            plot=False)

                rewards = eval_result['optim'].values.mean()

            return rewards

        try:
            study.optimize(objective,
                           n_trials=n_trials,
                           n_jobs=n_jobs,
                           timeout=timeout)
        except KeyboardInterrupt:
            logging.warning("Evaluation stopped.")

        # continue
        best_trial = study.best_trial

        if self.verbose > 0:
            print('Number of finished trials: ', len(study.trials))

            print('Best trial:')

            print('Value: ', best_trial.value)

            print('Params: ')
            for key, value in best_trial.params.items():
                print('    {}: {}'.format(key, value))

        # update using best parameters
        self.init_kwargs.update(best_trial.params)

        return best_trial, study.trials_dataframe()


#
# Aux functions
#


def _fit_worker(args):
    agent_class, train_env, init_kwargs, fit_kwargs = args
    agent = agent_class(train_env, copy_env=False,
                        reseed_env=False, **init_kwargs)
    info = agent.fit(**fit_kwargs)
    return agent, info



import numpy as np
import pytest
import rlberry.seeding as seeding

from copy import deepcopy
from rlberry.envs.classic_control import MountainCar, Acrobot
from rlberry.envs.finite import Chain
from rlberry.envs.finite import GridWorld
from rlberry.envs.benchmarks.ball_exploration import PBall2D, SimplePBallND


classes = [
    MountainCar,
    GridWorld,
    Chain,
    PBall2D,
    SimplePBallND,
    Acrobot
]


def get_env_trajectory(env, horizon):
    states = []
    ss = env.reset()
    for ii in range(horizon):
        states.append(ss)
        ss, _, _, _ = env.step(env.action_space.sample())
    return states


def compare_trajectories(traj1, traj2):
    for ss1, ss2 in zip(traj1, traj2):
        if not np.array_equal(ss1, ss2):
            return False
    return True


@pytest.mark.parametrize("ModelClass", classes)
def test_env_seeding(ModelClass):

    seeding.set_global_seed(123)
    env1 = ModelClass()

    seeding.set_global_seed(456)
    env2 = ModelClass()

    seeding.set_global_seed(123)
    env3 = ModelClass()

    if deepcopy(env1).is_online():
        traj1 = get_env_trajectory(env1, 500)
        traj2 = get_env_trajectory(env2, 500)
        traj3 = get_env_trajectory(env3, 500)

        assert not compare_trajectories(traj1, traj2)
        assert compare_trajectories(traj1, traj3)


@pytest.mark.parametrize("ModelClass", classes)
def test_copy_reseeding(ModelClass):

    seeding.set_global_seed(123)
    env = ModelClass()

    c_env = deepcopy(env)
    c_env.reseed()

    if deepcopy(env).is_online():
        traj1 = get_env_trajectory(env, 500)
        traj2 = get_env_trajectory(c_env, 500)
        assert not compare_trajectories(traj1, traj2)



import rlberry.seeding as seeding


def test_seeding():
    seed = 123
    seeding.set_global_seed(seed)
    assert seeding._GLOBAL_SEED_SEQ.entropy == seed

    _ = seeding.get_rng()
    assert seeding._GLOBAL_SEED_SEQ.n_children_spawned == 1

    _ = seeding.get_rng()
    assert seeding._GLOBAL_SEED_SEQ.n_children_spawned == 2


def test_random_numbers():
    seed = 43
    seeding.set_global_seed(seed)
    rng1 = seeding.get_rng()
    data1 = rng1.integers(100, size=1000)

    seed = 44
    seeding.set_global_seed(seed)
    rng2 = seeding.get_rng()
    data2 = rng2.integers(100, size=1000)

    seed = 44
    seeding.set_global_seed(seed)
    rng3 = seeding.get_rng()
    data3 = rng3.integers(100, size=1000)

    assert (data1 != data2).sum() > 5
    assert (data2 != data3).sum() == 0



from rlberry.agents.agent import Agent
from rlberry.agents.dynprog.utils import backward_induction, value_iteration
from rlberry.envs.finite.finite_mdp import FiniteMDP


class ValueIterationAgent(Agent):
    """
    Value iteration for enviroments of type FiniteMDP
    (rlberry.envs.finite.finite_mdp.FiniteMDP)

    Important: the discount gamma is also used if the problem is
    finite horizon, but, in this case, gamma can be set to 1.0.
    """

    name = "ValueIteration"
    fit_info = ("n_iterations", "precision")

    def __init__(self, env, gamma=0.95, horizon=None, epsilon=1e-6, **kwargs):
        """
        Parameters
        -----------
        env : rlberry.envs.finite.finite_mdp.FiniteMDP
        gamma : double
            discount factor in [0, 1]
        horizon : int
            horizon, if the problem is finite-horizon. if None, the discounted
            problem is solved
            default = None
        epsilon : double
            precision of value iteration, only used in discounted problems
            (when horizon is None).
        """
        # initialize base class
        assert isinstance(env, FiniteMDP), \
            "Value iteration requires a FiniteMDP model."
        Agent.__init__(self, env, **kwargs)

        #
        self.gamma = gamma
        self.horizon = horizon
        self.epsilon = epsilon

        # value functions
        self.Q = None
        self.V = None

    def fit(self, **kwargs):
        """
        Run value iteration.
        """
        info = {}
        if self.horizon is None:
            assert self.gamma < 1.0, \
                "The discounted setting requires gamma < 1.0"
            self.Q, self.V, n_it = value_iteration(self.env.R, self.env.P,
                                                   self.gamma, self.epsilon)
            info["n_iterations"] = n_it
            info["precision"] = self.epsilon
        else:
            self.Q, self.V = backward_induction(self.env.R, self.env.P,
                                                self.horizon, self.gamma)
            info["n_iterations"] = self.horizon
            info["precision"] = 0.0
        return info

    def policy(self, state, hh=0, **kwargs):
        """
        Parameters
        -----------
        state : int
        hh : int
            stage when action is taken (for finite horizon problems,
            the optimal policy depends on hh) not used if horizon is None.
        """
        assert self.env.observation_space.contains(state)
        if self.horizon is None:
            return self.Q[state, :].argmax()
        else:
            assert hh >= 0 and hh < self.horizon
            return self.Q[hh, state, :].argmax()



"""
Interface that allows 2D rendering.
"""

from abc import ABC, abstractmethod

from rlberry.rendering.opengl_render2d import OpenGLRender2D
from rlberry.rendering.pygame_render2d import PyGameRender2D
from rlberry.rendering.utils import video_write


class RenderInterface(ABC):
    """
    Common interface for rendering in rlberry.
    """

    def __init__(self):
        self._rendering_enabled = False
        self._debug_mode = False  # for testing and debug only

    def is_render_enabled(self):
        return self._rendering_enabled

    def enable_rendering(self):
        self._rendering_enabled = True

    def disable_rendering(self):
        self._rendering_enabled = False

    def save_video(self, filename, **kwargs):
        """
        Save video file.
        """
        pass

    @abstractmethod
    def render(self, **kwargs):
        """
        Display on screen.
        """
        pass


class RenderInterface2D(RenderInterface):
    """
    Interface for 2D rendering in rlberry.
    """

    def __init__(self):
        RenderInterface.__init__(self)
        self._rendering_enabled = False
        self._rendering_type = "2d"
        self._state_history_for_rendering = []
        self._refresh_interval = 50   # in milliseconds
        self._clipping_area = (-1.0, 1.0, -1.0, 1.0)  # (left,right,bottom,top)

        # rendering type, either 'pygame' or 'opengl'
        self.renderer_type = 'opengl'

    def get_renderer(self):
        if self.renderer_type == 'opengl':
            return OpenGLRender2D()
        elif self.renderer_type == 'pygame':
            return PyGameRender2D()
        else:
            raise NotImplementedError("Unknown renderer type.")

    @abstractmethod
    def get_scene(self, state):
        """
        Return scene (list of shapes) representing a given state
        """
        pass

    @abstractmethod
    def get_background(self):
        """
        Returne a scene (list of shapes) representing the background
        """
        pass

    def append_state_for_rendering(self, state):
        self._state_history_for_rendering.append(state)

    def set_refresh_interval(self, interval):
        self._refresh_interval = interval

    def clear_render_buffer(self):
        self._state_history_for_rendering = []

    def set_clipping_area(self, area):
        self._clipping_area = area

    def _get_background_and_scenes(self):
        # background
        background = self.get_background()

        # data: convert states to scenes
        scenes = []
        for state in self._state_history_for_rendering:
            scene = self.get_scene(state)
            scenes.append(scene)
        return background, scenes

    def render(self, **kwargs):
        """
        Function to render an environment that implements the interface.
        """

        if self.is_render_enabled():
            # background and data
            background, data = self._get_background_and_scenes()

            if len(data) == 0:
                print("No data to render.")
                return

            # render
            renderer = self.get_renderer()

            renderer.window_name = self.name
            renderer.set_refresh_interval(self._refresh_interval)
            renderer.set_clipping_area(self._clipping_area)
            renderer.set_data(data)
            renderer.set_background(background)
            renderer.run_graphics(self._debug_mode)
            return 0
        else:
            print("Rendering not enabled for the environment.")
            return 1

    def save_video(self, filename, framerate=25, **kwargs):

        # background and data
        background, data = self._get_background_and_scenes()

        if len(data) == 0:
            print("No data to save.")
            return

        # get video data from renderer
        renderer = self.get_renderer()
        renderer.window_name = self.name
        renderer.set_refresh_interval(self._refresh_interval)
        renderer.set_clipping_area(self._clipping_area)
        renderer.set_data(data)
        renderer.set_background(background)

        video_data = renderer.get_video_data()
        video_write(filename, video_data, framerate=framerate)
